/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * console.c -- Console management
 */

#include <string.h>
#include <bios.h>

#include "com32_private.h"

void (*bios_log_callback)(int level, const char *, ...) = NULL;

/*-- set_display_mode ----------------------------------------------------------
 *
 *      Set the display to either firmware native text mode, or VBE graphic
 *      mode. This function does nothing and returns successfully.
 *
 * Parameters
 *      IN mode: DISPLAY_MODE_NATIVE_TEXT or DISPLAY_MODE_VBE
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int set_display_mode(display_mode_t UNUSED_PARAM(mode))
{
   return ERR_SUCCESS;
}

/*-- com32_putc ----------------------------------------------------------------
 *
 *      Wrapper for the 'Write Character' COM32 service.
 *
 * Parameters
 *      IN c: the character to be printed.
 *----------------------------------------------------------------------------*/
static void com32_putc(uint8_t c)
{
   com32sys_t iregs;

   memset(&iregs, 0, sizeof (iregs));
   iregs.eax.b[1] = 0x02;
   iregs.edx.b[0] = c;
   intcall(COM32_INT_DOS_COMPATIBLE, &iregs, NULL);
}

/*-- firmware_print ------------------------------------------------------------
 *
 *      Print a string to the COM32 console.
 *
 * Parameters
 *      IN str: pointer to the ASCII string to print
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int firmware_print(const char *str)
{
   for ( ; *str != '\0'; str++) {
      if (*str == '\n') {
         com32_putc('\r');
      }
      com32_putc(*str);
   }

   return ERR_SUCCESS;
}

/*-- set_firmware_log_callback -------------------------------------------------
 *
 *      Set the routine to be called in order to log a message from the COM32
 *      layer.
 *
 * Parameters
 *      IN callback: pointer to the logging routine, or NULL to disable COM32
 *                   logging
 *----------------------------------------------------------------------------*/
void set_firmware_log_callback(void (*callback)(int, const char *, ...))
{
   bios_log_callback = callback;
}

/*-- get_serial_port -----------------------------------------------------------
 *
 *      Get the I/O base address of a COM serial port.
 *
 * Parameters
 *      IN  com:  1=COM1, 2=COM2, 3=COM3, 4=COM4
 *      OUT port: the serial port I/O base address
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_serial_port(int com, uint16_t *port)
{
   uint16_t p;

   if (com < 1 || com > 4) {
      return ERR_INVALID_PARAMETER;
   }

   p = bios_get_com_port(com);
   if (p == 0) {
      return ERR_UNSUPPORTED;
   }

   *port = p;

   return ERR_SUCCESS;
}
