/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * systab.c -- System table parsing
 */

#include <string.h>
#include <bios.h>
#include <mps.h>
#include <sm_bios.h>
#include <bootlib.h>

#include "com32_private.h"

typedef struct {
   uint32_t base;
   uint32_t len;
} addr32_range_t;

/*-- find_system_table ---------------------------------------------------------
 *
 *      Locate a system table in the given ranges of memory.
 *
 * Parameters
 *      IN  scan:        array of memory address ranges to be scanned
 *      IN  paragraph:   the table must be aligned on this boundary
 *      IN  size:        table minimum size
 *      IN  check_table: pointer to a function that checks whether a valid table
 *                       is present at a given address
 *      OUT table:       pointer to the table
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int find_system_table(addr32_range_t *scan, uint32_t paragraph,
                             size_t size, int (*check_table)(void *),
                             void **table)
{
   uint32_t len;
   char *p;

   while (scan->len > 0) {
      len = scan->len;

      if (len > paragraph) {
         p = UINT64_TO_PTR(roundup64(scan->base, paragraph));
         len -= (ptrdiff_t)p - (ptrdiff_t)scan->base;

         while (len >= size) {
            if (check_table(p) == ERR_SUCCESS) {
               *table = p;
               return ERR_SUCCESS;
            }

            p += paragraph;
            len -= paragraph;
         }
      }

      scan++;
   }

   return ERR_NOT_FOUND;
}

/*-- smbios_check_eps ----------------------------------------------------------
 *
 *      Check the SMBIOS Entry Point Structure (EPS) integrity:
 *        - table signature is correct
 *        - RSDP checksum(s) make the structure bytes add up to zero
 *
 * Parameters
 *      IN eps: pointer to the EPS
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int smbios_check_eps(void *eps)
{
   SMBIOS_entry_point_t *p = eps;

   if (memcmp(p->anchorString, SMBIOS_EPS_SIGNATURE,
              SMBIOS_EPS_SIGNATURE_LEN) != 0) {
      return ERR_NOT_FOUND;
   }

   if (!is_valid_bios_table(p, p->length)) {
      return ERR_CRC_ERROR;
   }

   return ERR_SUCCESS;
}

/*-- get_smbios_table ----------------------------------------------------------
 *
 *      Get the SMBIOS Entry Point Structure (EPS).
 *
 *      On non-EFI systems, the SMBIOS Entry Point structure, described below,
 *      can be located by application software by searching for the anchor-
 *      string on paragraph (16-byte) boundaries within the physical memory
 *      address range 000F0000h to 000FFFFFh.
 *
 * Parameters
 *      OUT eps: pointer to the EPS
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_smbios_table(void **eps)
{
   addr32_range_t scan[2] = {
      {0xf0000, BIOS_UPPER_MEM_START - 0xf0000},
      {0, 0}
   };

   return find_system_table(scan, SMBIOS_PARAGRAPH_SIZE,
                            sizeof (SMBIOS_entry_point_t),
                            smbios_check_eps, eps);
}

/*-- mps_check_float_ptr -------------------------------------------------------
 *
 *      Check the MP Floating Pointer integrity:
 *        - table signature is correct
 *        - RSDP checksum(s) make the structure bytes add up to zero
 *
 * Parameters
 *      IN fltptr: pointer to the MPS float pointer
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int mps_check_float_ptr(void *fltptr)
{
   MPFloatingPointer_t *p = fltptr;

   if (memcmp(p->signature, MP_FLOAT_PTR_SIGNATURE,
              MP_FLOAT_PTR_SIGNATURE_LEN) != 0) {
      return ERR_NOT_FOUND;
   }

   if (!is_valid_bios_table(p, p->length * MPS_PARAGRAPH_SIZE)) {
      return ERR_CRC_ERROR;
   }

   return ERR_SUCCESS;
}

/*-- get_mps_table -------------------------------------------------------------
 *
 *      Get the MP floating pointer.
 *
 *      The operating system searches for the MP floating pointer structure in
 *      the order described below:
 *        a. In the first kilobyte of Extended BIOS Data Area (EBDA), or
 *        b. Within the last kilobyte of system base memory (e.g., 639K-640K for
 *            systems with 640KB of base memory or 511K-512K for systems with
 *            512 KB of base memory) if the EBDA segment is undefined, or
 *        c. In the BIOS ROM address space between 0F0000h and 0FFFFFh.
 *
 * Parameters
 *      OUT mps: pointer to the MP floating pointer.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_mps_table(void **fltptr)
{
   int i;
   addr32_range_t scan[4] = {
      {0, 1024},
      {LOWMEM_LIMIT - 1024, 1024},
      {0xf0000, BIOS_UPPER_MEM_START - 0xf0000},
      {0, 0}
   };

   scan[0].base = bios_get_ebda();
   i = (scan[0].base != 0) ? 0 : 1;

   return find_system_table(&scan[i], MPS_PARAGRAPH_SIZE,
                            sizeof (MPFloatingPointer_t),
                            mps_check_float_ptr, fltptr);
}
