/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * e820.c -- E820 memory map management functions
 */

#include <string.h>
#include <e820.h>
#include <bootlib.h>

#define IS_E820_MERGEABLE(r1, r2)                                       \
   ((r1)->type == (r2)->type &&                                         \
    (r1)->attributes == (r2)->attributes &&                             \
    is_mergeable(E820_BASE(r1), E820_LENGTH(r1),                        \
                 E820_BASE(r2), E820_LENGTH(r2)))

/*-- is_mergeable --------------------------------------------------------------
 *
 *      Check whether two integer ranges can be merged. Two ranges can be merged
 *      if they overlap, or if they are contiguous.
 *
 * Parameters
 *      IN a1: first range start
 *      IN l1: first range length
 *      IN a2: second range start
 *      IN l2: second range length
 *
 * Results
 *      true/false
 *----------------------------------------------------------------------------*/
bool is_mergeable(uint64_t a1, uint64_t l1, uint64_t a2, uint64_t l2)
{
   uint64_t tmp;

   if (a2 < a1) {
      tmp = a1;
      a1 = a2;
      a2 = tmp;
      l1 = l2;
   }

   if (a1 + l1 < a1) {
      return true;
   }

   return a2 <= a1 + l1;
}

/*-- is_overlap ----------------------------------------------------------------
 *
 *      Check whether two integer ranges overlap.
 *
 * Parameters
 *      IN a1: first range start
 *      IN l1: first range length
 *      IN a2: second range start
 *      IN l2: second range length
 *
 * Results
 *      true/false
 *----------------------------------------------------------------------------*/
bool is_overlap(uint64_t a1, uint64_t l1, uint64_t a2, uint64_t l2)
{
   uint64_t tmp;

   if (a2 < a1) {
      tmp = a1;
      a1 = a2;
      a2 = tmp;
      l1 = l2;
   }

   if (a1 + l1 < a1) {
      return true;
   }

   return a2 < a1 + l1;
}

/*-- e820_compare --------------------------------------------------------------
 *
 *      For sorting the memory map descriptors by increasing base addresses.
 *
 * Parameters
 *      IN a: pointer to a memory map descriptor
 *      IN b: pointer to another memory map descriptor
 *
 * Results
 *      -1 when a < b
 *       1 when a > b
 *       0 when a == b
 *----------------------------------------------------------------------------*/
static int e820_compare(const void *a, const void *b)
{
   if (E820_BASE((const e820_range_t *)a) < E820_BASE((const e820_range_t *)b)) {
      return -1;
   }
   if (E820_BASE((const e820_range_t *)a) > E820_BASE((const e820_range_t *)b)) {
      return 1;
   }
   return 0;
}

/*-- e820_mmap_merge -----------------------------------------------------------
 *
 *      Merge memory map descriptors when they report contiguous memory of the
 *      same type.
 *
 * Parameters
 *      IN  mmap:  pointer to the memory map
 *      IN  count: original number of entries in the memory map
 *      OUT count: updated number of entries in the memory map
 *----------------------------------------------------------------------------*/
void e820_mmap_merge(e820_range_t *mmap, size_t *count)
{
   size_t idx, n;
   uint64_t len;

   if (*count > 1) {
      bubble_sort(mmap, *count, sizeof (e820_range_t), e820_compare);

      for (idx = 0; idx < *count - 1; idx++) {
         for (n = 0; n < *count - idx - 1; n++) {
            if (!IS_E820_MERGEABLE(&mmap[n], &mmap[n + 1])) {
               break;
            }
         }

         if (n > 0) {
            len = E820_BASE(&mmap[n]) + E820_LENGTH(&mmap[n]) - E820_BASE(mmap);
            e820_set_entry(mmap, E820_BASE(mmap), len, mmap->type,
                           mmap->attributes);

            if (idx + n + 1 < *count) {
               memmove(&mmap[1], &mmap[n + 1],
                       (*count - (idx + n + 1)) * sizeof (e820_range_t));
            }
            *count -= n;
         }

         mmap++;
      }
   }
}
