/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *****************************************************************************/

/*
 * serial.c -- Serial console
 */

#include <stdint.h>
#include <boot_services.h>
#include <bootlib.h>

#define UART_RX               0         /* In:  Receive buffer (DLAB=0) */
#define UART_TX               0         /* Out: Transmit buffer (DLAB=0) */
#define UART_DLL              0         /* Out: Divisor Latch Low (DLAB=1) */
#define UART_DLM              1         /* Out: Divisor Latch High (DLAB=1) */
#define UART_IER              1         /* Out: Interrupt Enable Register */
#define UART_FCR              2         /* Out: FIFO Control Register */
#define UART_LCR              3         /* Out: Line Control Register */
#define UART_MCR              4         /* Out: Modem Control Register */
#define UART_LSR              5         /* In:  Line Status Register */
#define UART_MSR              6         /* In:  Modem Status Register */

#define UART_FCR_ENABLE_FIFO  (1 << 0)  /* Enable the FIFO */
#define UART_FCR_CLEAR_RCVR   (1 << 1)  /* Clear the RCVR FIFO */
#define UART_FCR_CLEAR_XMIT   (1 << 2)  /* Clear the XMIT FIFO */
#define UART_FCR_TRIGGER_1    0x00      /* Mask for trigger set at 1 */

#define UART_LCR_WLEN8        0x03      /* Word length: 8 bits */
#define UART_LCR_SBC          (1 << 6)  /* Set break control */
#define UART_LCR_DLAB         (1 << 7)  /* Divisor latch access bit */

#define UART_MCR_DTR          (1 << 0)  /* DTR complement */
#define UART_MCR_RTS          (1 << 1)  /* RTS complement */
#define UART_MCR_OUT2         (1 << 3)  /* Out2 complement */

#define UART_MSR_CTS          (1 << 4)  /* Clear to Send */
#define UART_MSR_DCD          (1 << 7)  /* Data Carrier Detect */

#define UART_LSR_THRE         (1 << 5)  /* Transmit-hold-register empty */

static INLINE void io_delay(void)
{
   __asm__ __volatile__ ("outb %al, $0x80");
}

static INLINE uint8_t inb(uint16_t port)
{
   uint8_t val;

   __asm__ __volatile__ ("inb %w1, %b0"
                         : "=a" (val)
                         : "Nd" (port));

   return val;
}

static INLINE void outb(uint16_t port, uint8_t val)
{
   __asm__ __volatile__ ("outb %b1, %w0"
                         :
                         : "Nd" (port), "a" (val));
   io_delay();
}

static uint16_t serial_port = 0;

/*-- serial_putc ---------------------------------------------------------------
 *
 *      Write a character on a serial port.
 *
 * Parameters
 *      IN port: port to write to
 *      IN c:    character to be written
 *----------------------------------------------------------------------------*/
static void serial_putc(uint16_t port, char c)
{
   uint16_t timeout;

   for (timeout = 0xffff; timeout > 0; timeout--) {
      if ((inb(port + UART_LSR) & UART_LSR_THRE) != 0) {
         outb(port + UART_TX, c);
         return;
      }
   }
}

/*-- serial_init ---------------------------------------------------------------
 *
 *      Initialize a serial port with the given baudrate.
 *        - polling mode (no interrupts)
 *        - Word length = 8 bits
 *        - 1 stop bit
 *        - no parity
 *        - FIFO triggering on 1 byte
 *
 * Parameters
 *      IN port:     serial port
 *      IN baudrate: baud rate, in bits per second
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int serial_init(uint16_t port, uint32_t baudrate)
{
   uint16_t latch;
   uint8_t c;

   if (baudrate == 0) {
      return ERR_INVALID_PARAMETER;
   }

   /* Word length = 8, no parity, 1 stop bit */
   outb(port + UART_LCR, UART_LCR_WLEN8);
   c = inb(port + UART_LCR);
   if (c != UART_LCR_WLEN8) {
      return ERR_DEVICE_ERROR;
   }

   /* No interrupts */
   outb(port + UART_IER, 0);

   /*
    * Some null modem cables use a loop-back from the DTR to the DCD line and
    * another from RTS to CTS.
    */
   outb(port + UART_MCR, UART_MCR_OUT2 | UART_MCR_RTS | UART_MCR_DTR);

   /* Set baudrate */
   latch = MAX(115200 / baudrate, 1);
   outb(port + UART_LCR, c | UART_LCR_DLAB);
   outb(port + UART_DLL, latch & 0xff);
   outb(port + UART_DLM, latch >> 8);
   outb(port + UART_LCR, c & ~UART_LCR_DLAB);

   /* Try to enable the FIFO (trigger on 1 bytes) */
   outb(port + UART_FCR, UART_FCR_ENABLE_FIFO);
   c = inb(port + UART_FCR);
   if (c == UART_FCR_ENABLE_FIFO) {
      c |= UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT | UART_FCR_TRIGGER_1;
   } else {
      c = 0;
   }
   outb(port + UART_FCR, c);

   /* Read the line status register to clear the error flags */
   c = inb(port + UART_LSR);

   return ERR_SUCCESS;
}

/*-- serial_log ----------------------------------------------------------------
 *
 *      Write a string to the serial console.
 *
 * Parameters
 *      IN msg: pointer to the string to be written
 *
 * Results
 *      The number of characters actually written.
 *----------------------------------------------------------------------------*/
static int serial_log(const char *msg)
{
   int len, level;
   bool newline;

   newline = true;
   len = 0;

   for ( ; *msg != '\0'; msg++) {
      if (newline && syslog_get_message_level(msg, &level) == ERR_SUCCESS) {
         msg += 3;
         if (*msg == '\0') {
            break;
         }
      }

      if (*msg == '\n') {
         serial_putc(serial_port, '\r');
         len++;
         newline = true;
      } else {
         newline = false;
      }

      serial_putc(serial_port, *msg);
      len++;
   }

   return len;
}

/*-- serial_log_init -----------------------------------------------------------
 *
 *      Initialize the serial console.
 *
 * Parameters
 *      IN com:      serial port COM number (1=COM1, 2=COM2, 3=COM3, 4=COM4)
 *      IN baudrate: serial speed, in bits per second
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int serial_log_init(int com, uint32_t baudrate)
{
   uint16_t port;
   int status;

   if (baudrate == 0 || com < 1 || com > 4) {
      return ERR_INVALID_PARAMETER;
   }

   status = get_serial_port(com, &port);
   if (status != ERR_SUCCESS) {
      return status;
   }

   status = serial_init(port, baudrate);
   if (status != ERR_SUCCESS) {
      return status;
   }

   status = log_subscribe(serial_log, LOG_DEBUG);
   if (status != ERR_SUCCESS) {
      return status;
   }

   serial_port = port;

   return ERR_SUCCESS;
}
