/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * bootlib.h -- High level (platform independent) boot utils routines
 */

#ifndef BOOTLIB_H_
#define BOOTLIB_H_

#include <compat.h>
#include <sys/types.h>
#include <stdbool.h>
#include <stdarg.h>
#include <string.h>
#include <fb.h>
#include <syslog.h>
#include <e820.h>
#include <disk.h>
#include <mbr.h>

#define ceil(x,y) (((x) + (y) - 1) / (y))

typedef struct {
   mbr_part_t info;
   int id;
} partition_t;

/*
 * e820.c
 */

EXTERN void e820_mmap_merge(e820_range_t *mmap, size_t *count);
EXTERN bool is_mergeable(uint64_t a1, uint64_t l1, uint64_t a2, uint64_t l2);
EXTERN bool is_overlap(uint64_t a1, uint64_t l1, uint64_t a2, uint64_t l2);

/*
 * string.c
 */
static INLINE bool is_absolute(const char *filepath)
{
   return (filepath[0] == '/' || filepath[0] == '\\' ||
           strstr(filepath, "://") != NULL);
}

EXTERN int str_alloc(size_t length, char **str);
EXTERN void mem_swap(void *m1, void *m2, size_t n);
EXTERN char *mem_strcasestr(const void *src, const char *str, size_t n);
EXTERN char *str_merge_spaces(char *str);
EXTERN int str_to_argv(char *cmdline, int *argc, char ***argv);
EXTERN int argv_to_str(int argc, char **argv, char **s);
EXTERN bool is_number(const char *str);
EXTERN int make_path(const char *root, const char *filepath, char **buffer);
EXTERN int insert_char(char *strbuf, char c, size_t n);
EXTERN int delete_char(char *str, size_t n);

/*
 * gzip.c
 */
EXTERN bool is_gzip(const void *buffer, size_t size);
EXTERN int gzip_extract(const void *src, size_t src_size, void **dest,
                        size_t *dest_size);

/*
 * file.c
 */
#define MAX_PATH_LEN 2048

EXTERN int file_get_size_hint(int volid, const char *filename,
                              size_t *filesize);
EXTERN int file_load(int volid, const char *filename, int (*callback)(size_t),
                     void **buffer, size_t *bufsize);
EXTERN int file_overwrite(int volid, const char *filepath, void *buffer,
                          size_t size);

/*
 * volume.c
 */
#define FIRMWARE_BOOT_VOLUME 0
EXTERN int volume_read(int volid, void *dest, uint64_t offset, size_t size);
EXTERN int get_volume_info(disk_t *disk, int part_id, partition_t *partition);

/*
 * mbr.c
 */
EXTERN int mbr_get_part_info(disk_t *disk, char *mbr, int part_id,
                             partition_t *partition);

/*
 * gpt.c
 */
EXTERN int gpt_get_part_info(disk_t *disk, int part_id, partition_t *partition);

/*
 * log.c
 */
#define IS_SYSLOG_LEVEL(_level_)                                        \
   ((_level_) >= 0 && (_level_) <= LOG_DEBUG)

#define IS_SYSLOG_LEVEL_STR(_level_) IS_SYSLOG_LEVEL((_level_) - '0')

static INLINE bool is_syslog_message(const char *str)
{
   return str[0] == '<' && IS_SYSLOG_LEVEL_STR(str[1]) && str[2] == '>';
}

typedef int (*log_callback_t)(const char *msg);

EXTERN void Log(int level, const char *fmt, ...);
EXTERN int log_subscribe(log_callback_t callback, int maxlevel);
EXTERN void log_unsubscribe(log_callback_t callback);
EXTERN const char *log_buffer_addr(void);
EXTERN int log_init(void);
EXTERN int syslog_get_message_level(const char *msg, int *level);

/*
 * Serial port
 */
EXTERN int serial_log_init(int com, uint32_t baudrate);

/*
 * sort.c
 */
EXTERN void bubble_sort(void *base, size_t nmemb, size_t size,
                        int (*compar)(const void *, const void *));

/*
 * fbcon.c
 */
EXTERN int fbcon_init(framebuffer_t *fbinfo, font_t *cons_font, int x, int y,
                      unsigned int width, unsigned int height);
EXTERN void fbcon_reset(void);
EXTERN void fbcon_clear(void);
EXTERN void fbcon_shutdown(void);

/*
 * parse.c
 */
typedef enum {
   OPT_STRING,
   OPT_INTEGER,
   OPT_INVAL
} option_type_t;

typedef union {
   char *str;              /* For storing an OPT_STRING option */
   int integer;            /* For storing an OPT_INTEGER option */
} option_value_t;

typedef struct {
   const char *key;
   const char *separator;
   option_value_t value;
   option_type_t type;
} option_t;

EXTERN int parse_config_file(int volid, const char *filename,
                             option_t *options);

/*
 * video.c
 */
typedef struct {
   vbe_t controller;            /* VBE controller info */
   vbe_mode_t mode;             /* Current VBE mode info */
   vbe_mode_id_t *modes_list;   /* Supported VBE modes list */
   vbe_mode_id_t current_mode;  /* Current VBE mode ID */
} vbe_info_t;

EXTERN int video_check_support(void);
EXTERN int video_get_vbe_info(vbe_info_t *vbe_info);
EXTERN int video_set_mode(framebuffer_t *fb, unsigned int width,
                          unsigned int height, unsigned int depth,
                          bool allow_smaller);
EXTERN int video_set_text_mode(void);

/*-- roundup64 ----------------------------------------------------------------
 *
 *      Round n up to the nearest multiple of unit.
 *
 * Parameters
 *      IN n:    input value
 *      IN unit: round up to a multiple of this much
 *
 * Results
 *      The rounded value.
 *----------------------------------------------------------------------------*/
static INLINE uint64_t roundup64(uint64_t n, uint64_t unit)
{
   uint64_t u;

   u = n / unit;
   if (n % unit > 0) {
      u++;
   }

   return u * unit;
}

/*
 *  For casting safely between 32-bit and 64-bit values
 */
#define PTR_TO_UINT(_ptr_)    ((uintptr_t)(_ptr_))
#define PTR_TO_UINT64(_ptr_)  ((uint64_t)PTR_TO_UINT(_ptr_))
#define UINT_TO_PTR(_ui_)     ((void *)(uintptr_t)(_ui_))
#define UINT32_TO_PTR(_ui32_) UINT_TO_PTR(_ui32_)

#define IS_UINT32(_ui32_)     (((uint64_t)(_ui32_) & 0xffffffff00000000) == 0)
#define IS_PTR32(_ptr_)       IS_UINT32(PTR_TO_UINT(_ptr_))
#define IS_REGION32(_ptr_, _size_)                                \
  (IS_PTR32(_ptr_) && IS_UINT32(PTR_TO_UINT(_ptr_) + (_size_)))

#define PTR_TO_UINT32(_ptr_)  ((uint32_t)PTR_TO_UINT(_ptr_))
#define UINT64_TO_PTR(_ui64_) UINT_TO_PTR(_ui64_)

#endif /* !BOOTLIB_H_ */
