/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * compat.h -- Portability across building environments
 */

#ifndef COMPAT_H_
#define COMPAT_H_

#include <compiler.h>
#include <error.h>
#include <sys/types.h>

/*
 * Debugging
 */
#ifdef DEBUG
#   define ASSERT_GENERIC(_expr_, _callback_, _msg_)              \
       do {                                                       \
          if (!(_expr_)) {                                        \
             (_callback_)(__FILE__ " (%d): "                      \
                          "assert (" _msg_ ")\n", __LINE__);      \
             for ( ; ; );                                         \
          }                                                       \
       } while (0)
#else
#   define ASSERT_GENERIC(_expr_, _callback_, _msg_)
#endif

/*
 * Various usual macro
 */
#define IS_ALIGNED(_addr_, _alignment_)                           \
   (((_addr_) % (_alignment_)) == 0)

#define STRSIZE(_str_)                                            \
   (strlen(_str_) + 1)

#ifndef ARRAYSIZE
#   define ARRAYSIZE(_array_) (sizeof (_array_) / sizeof ((_array_)[0]))
#endif

#if !defined(MIN)
#   define MIN(a,b)  (((a)<(b))?(a):(b))
#endif
#if !defined(MAX)
#   define MAX(a,b)  (((a)>(b))?(a):(b))
#endif

/*
 * The main purpose of the uint64_generic_t type and the derived inline
 * functions is to bring up portability across 32-bit compilers that do not
 * support shifting, multiplying and dividing 64-bit integers.
 *
 * On these compilers, isolating the 4 MSB of a 64-bit value is not possible
 * using the trivial expression (ui64 >> 32).
 */
#pragma pack(1)
typedef union {
   struct {
      uint32_t low;
      uint32_t high;
   } ui32;

   uint64_t ui64;
} uint64_generic_t;
#pragma pack()

static INLINE uint64_t uint32_concat(uint32_t high, uint32_t low)
{
   uint64_generic_t llu;

   llu.ui32.low = low;
   llu.ui32.high = high;

   return llu.ui64;
}

static INLINE uint32_t highhalf64(uint64_t u)
{
   uint64_generic_t llu;

   llu.ui64 = u;

   return llu.ui32.high;
}

static INLINE uint32_t lowhalf64(uint64_t u)
{
   uint64_generic_t llu;

   llu.ui64 = u;

   return llu.ui32.low;
}

#endif /* COMPAT_H_ */
