/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * elf.c -- ELF32/64 parsing functions
 */

#include <elf.h>
#include "mboot.h"

/*-- is_x86_elf ----------------------------------------------------------------
 *
 *      ELF32/64 sanity checks for the x86 architecture.
 *
 * Parameters
 *      IN buffer: pointer to the ELF binary buffer
 *      IN buflen: size of the buffer, in bytes
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int is_x86_elf(void *buffer, size_t buflen)
{
   Elf32_Ehdr *ehdr32 = buffer;
   Elf_CommonEhdr *ehdr = buffer;
   bool is64;

   if (buflen < EI_NIDENT + 4 || !IS_ELF(*ehdr32)) {
      return ERR_BAD_TYPE;
   }

   if (ehdr32->e_ident[EI_CLASS] != ELFCLASS32 &&
       ehdr32->e_ident[EI_CLASS] != ELFCLASS64) {
      return ERR_BAD_ARCH;
   }

   if ((Elf_CommonEhdrGetMachine(ehdr) != EM_386 &&
        Elf_CommonEhdrGetMachine(ehdr) != EM_X86_64) ||
       ehdr32->e_ident[EI_DATA] != ELFDATA2LSB) {
      return ERR_BAD_ARCH;
   }

   is64 = Elf_CommonEhdrIs64(ehdr);

   if (buflen < Elf_CommonEhdrSize(is64)) {
      return ERR_UNEXPECTED_EOF;
   }

   if (Elf_CommonEhdrGetPhNum(ehdr) <= 0) {
      return ERR_BAD_HEADER;
   }

   if ((size_t)Elf_CommonEhdrGetPhNum(ehdr) * Elf_CommonPhdrSize(is64)
       + Elf_CommonEhdrGetPhOff(ehdr) > buflen) {
      return ERR_UNEXPECTED_EOF;
   }

   if (ehdr32->e_ident[EI_VERSION] != EV_CURRENT ||
       Elf_CommonEhdrGetVersion(ehdr) != EV_CURRENT) {
      return WARNING(ERR_INCOMPATIBLE_VERSION);
   }

   if (Elf_CommonEhdrGetType(ehdr) != ET_EXEC) {
      return WARNING(ERR_NOT_EXECUTABLE);
   }

   if (Elf_CommonEhdrGetPhEntSize(ehdr) != Elf_CommonPhdrSize(is64)) {
      return WARNING(ERR_BAD_HEADER);
   }

   return ERR_SUCCESS;
}

/*-- elf_check_headers ---------------------------------------------------------
 *
 *      Parse ELF32/64 headers and make sure no segment overruns end of file.
 *
 * Parameters
 *      IN buffer: pointer to the ELF binary buffer
 *      IN buflen: size of the buffer, in bytes
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int elf_check_headers(void *buffer, size_t buflen)
{
   Elf_CommonEhdr *ehdr = buffer;
   Elf_CommonPhdr *phdr, *ph;
   Elf64_Size load_size, run_size;
   char *load_addr;
   unsigned int i;
   int status;

   status = is_x86_elf(buffer, buflen);
   if (status != ERR_SUCCESS && !IS_WARNING(status)) {
      return status;
   }

   phdr = (Elf_CommonPhdr *)((char *)buffer + Elf_CommonEhdrGetPhOff(ehdr));

   for (i = 0; i < Elf_CommonEhdrGetPhNum(ehdr); i++) {
      ph = Elf_CommonEhdrGetPhdr(ehdr, phdr, i);

      if (Elf_CommonPhdrGetType(ehdr, ph) == PT_LOAD) {
         load_addr = (char *)buffer + Elf_CommonPhdrGetOffset(ehdr, ph);
         load_size = Elf_CommonPhdrGetFilesz(ehdr, ph);
         run_size = Elf_CommonPhdrGetMemsz(ehdr, ph);

         if (load_addr + load_size > (char *)buffer + buflen) {
            return ERR_UNEXPECTED_EOF;
         }
      }
   }

   return status;
}

/*-- elf_register --------------------------------------------------------------
 *
 *      List the ELF32/64 sections.
 *
 * Parameters
 *      IN  buffer: pointer to the ELF binary buffer
 *      OUT entry:  ELF entry point
 *      OUT shdr:   pointer to the ELF section table header
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int elf_register(void *buffer, Elf_CommonAddr *entry, Elf_CommonShdr **shdr)
{
   char *load_addr;
   Elf_CommonAddr run_addr;
   Elf64_Size load_size, run_size, bss_size;
   Elf_CommonEhdr *ehdr = buffer;
   Elf_CommonPhdr *phdr, *ph;
   unsigned int i;
   int status;

   phdr = (Elf_CommonPhdr *)((char *)buffer + Elf_CommonEhdrGetPhOff(ehdr));

   for (i = 0; i < Elf_CommonEhdrGetPhNum(ehdr); i++) {
      /* Program header parsing */
      ph = Elf_CommonEhdrGetPhdr(ehdr, phdr, i);

      if (Elf_CommonPhdrGetType(ehdr, ph) == PT_LOAD) {
         load_addr = (char *)buffer + Elf_CommonPhdrGetOffset(ehdr, ph);
         load_size = Elf_CommonPhdrGetFilesz(ehdr, ph);
         run_addr = Elf_CommonPhdrGetPaddr(ehdr, ph);
         run_size = Elf_CommonPhdrGetMemsz(ehdr, ph);
         bss_size = run_size - load_size;

         /* Set up the segment */
         if (run_size - bss_size > 0) {
            status = add_kernel_object(load_addr, run_size - bss_size, run_addr);
            if (status != ERR_SUCCESS) {
               return status;
            }
         }

         /* Set up the BSS */
         if (bss_size > 0) {
            status = add_kernel_object(NULL, bss_size, run_addr + load_size);
            if (status != ERR_SUCCESS) {
               return status;
            }
         }
      }
   }

   if (Elf_CommonEhdrGetShOff(ehdr) > 0) {
      /* Section header parsing */
      load_addr = (char *)ehdr + Elf_CommonEhdrGetShOff(ehdr);
      load_size = (Elf64_Size)Elf_CommonEhdrGetShEntSize(ehdr) *
                  (Elf64_Size)Elf_CommonEhdrGetShNum(ehdr);
      status = add_sysinfo_object(load_addr, load_size, ALIGN_PAGE);
      if (status != ERR_SUCCESS) {
         return status;
      }

      for (i = 0; i < Elf_CommonEhdrGetShNum(ehdr); i++) {
         if (!Elf_CommonShdrGetAddr(ehdr, i) && Elf_CommonShdrGetSize(ehdr, i)) {
            load_addr = (char *)buffer + Elf_CommonShdrGetOff(ehdr, i);
            load_size = Elf_CommonShdrGetSize(ehdr, i);
            status = add_sysinfo_object(load_addr, load_size,
                                        (size_t)Elf_CommonShdrGetAddrAlign(ehdr, i));
            if (status != ERR_SUCCESS) {
               return status;
            }
         }
      }

      *shdr = (Elf_CommonShdr *)((char *)ehdr + Elf_CommonEhdrGetShOff(ehdr));
   } else {
      *shdr = NULL;
   }

   *entry = Elf_CommonEhdrGetEntry(buffer);

   return ERR_SUCCESS;
}
