/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * load.c -- Kernel/modules loading
 */

#include <string.h>
#include <boot_services.h>
#include "mboot.h"

static void load_sanity_check(void)
{
   uint64_t load_size, offset;
   const char *filename, *msg;
   bool error, is_previous_loaded, is_loaded;
   unsigned int i;
   void *addr;

   error = false;

   if (boot.modules_nr == 0 || boot.modules == NULL) {
      error = true;
      Log(LOG_ERR, "Nothing to load.\n");
   } else {
      is_previous_loaded = true;
      offset = 0;

      for (i = 0; i < boot.modules_nr; i++) {
         msg = NULL;
         filename = boot.modules[i].filename;
         addr = boot.modules[i].addr;
         load_size = boot.modules[i].load_size;
         is_loaded = boot.modules[i].is_loaded;

         if (filename == NULL) {
            msg = "Module has no filename";
         } else if (!is_previous_loaded && is_loaded) {
            msg = "Previous module has been skipped";
         }

         if (msg != NULL) {
            error = true;
            Log(LOG_ERR, "Mod[%u]: %s.\n", i, msg);
         }

         is_previous_loaded = is_loaded;
         offset += load_size;
      }

      if (offset != boot.load_offset) {
         error = true;
         Log(LOG_ERR, "Inconsistent loading offset.\n");
      }
   }

   if (error) {
      Log(LOG_ERR, "Modules are corrupted.\n");
      while (1);
   }
}

/*-- load_callback -------------------------------------------------------------
 *
 *      Increment the load offset with a given amount of freshly loaded memory.
 *      This function is a callback for the file_load() function.
 *
 * Parameters
 *      IN chunk_size: amount of loaded memory, in bytes, since the last call to
 *                     this function
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int load_callback(size_t chunk_size)
{
   unsigned int i;

   for (i = 0; i < boot.modules_nr; i++) {
      if (!boot.modules[i].is_loaded) {
         boot.modules[i].load_size += chunk_size;
         break;
      }
   }

   boot.load_offset += chunk_size;

   gui_refresh();

   return ERR_SUCCESS;
}

/*-- get_module_size -----------------------------------------------------------
 *
 *      Get the size of a module.
 *
 *      In some circumstances, it is not possible to determine the full size of
 *      a file without loading it. If that is the case, then this function
 *      will return ERR_UNSUPPORTED.
 *
 * Parameters
 *      IN  n:    module id
 *      OUT size: the module size, in bytes
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int get_module_size(unsigned int n, size_t *size)
{
   const char *filepath;
   size_t filesize;
   int status;

  filepath = boot.modules[n].filename;

   status = file_get_size_hint(boot.volid, filepath, &filesize);
   if (status != ERR_SUCCESS) {
      if (status == ERR_NOT_FOUND) {
         Log(LOG_ERR, "%s: file not found.\n", filepath);
      } else if (status != ERR_UNSUPPORTED) {
         Log(LOG_ERR, "%s: unknown file size.\n", filepath);
      }

      return status;
   }

   *size = filesize;

   return ERR_SUCCESS;
}

/*-- get_load_size_hint --------------------------------------------------------
 *
 *      Get the total size of the data to be loaded.
 *
 *      In some circumstances, it is not possible to determine the full size of
 *      a file without loading it. If that is the case, then this function
 *      will return ERR_UNSUPPORTED.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_load_size_hint(void)
{
   size_t filesize;
   uint64_t bytes;
   unsigned int i;
   int status;

   boot.load_size = 0;
   bytes = 0;

   for (i = 0; i < boot.modules_nr; i++) {
      status = get_module_size(i, &filesize);
      if (status != ERR_SUCCESS) {
         return status;
      }

      bytes += filesize;
   }

   boot.load_size = bytes;

   return ERR_SUCCESS;
}

/*-- unload_boot_modules -------------------------------------------------------
 *
 *      Unload previously loaded boot modules.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
void unload_boot_modules(void)
{
   unsigned int i;

   memset(&boot.kernel, 0, sizeof (kernel_t));

   for (i = 0; i < boot.modules_nr; i++) {
      sys_free(boot.modules[i].addr);
      boot.modules[i].addr = NULL;
      boot.modules[i].load_size = 0;
      boot.modules[i].size = 0;
      boot.modules[i].is_loaded = false;
   }
}

/*-- load_module ---------------------------------------------------------------
 *
 *      Load a boot module.
 *
 * Parameters
 *      IN n: module id
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int load_module(unsigned int n)
{
   const char *filepath;
   size_t size;
   void *addr;
   int status;

   filepath = boot.modules[n].filename;
   Log(LOG_INFO, "Loading %s\n", filepath);

   status = file_load(boot.volid, filepath,
                      (boot.load_size > 0) ? load_callback : NULL,
                      &addr, &size);
   if (status != ERR_SUCCESS) {
      if (status != ERR_ABORTED) {
         Log(LOG_ERR, "Error loading %s\n", filepath);
      }

      return status;
   }

   if (n == 0) {
      status = check_multiboot_kernel(addr, size);
      if (status != ERR_SUCCESS) {
         Log(LOG_ERR, "%s: kernel is either invalid or corrupted.\n", filepath);
         sys_free(addr);
         return status;
      }
   }

   boot.modules[n].addr = addr;
   boot.modules[n].size = size;
   boot.modules[n].is_loaded = true;

   if (boot.load_size == 0) {
      status = load_callback(size);
      if (status != ERR_SUCCESS) {
         boot.modules[n].addr = NULL;
         boot.modules[n].size = 0;
         boot.modules[n].is_loaded = false;
         sys_free(addr);
         return status;
      }
   }

   return ERR_SUCCESS;
}

/*-- load_boot_modules----------------------------------------------------------
 *
 *      Load kernel and modules into memory (do not relocate them).
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int load_boot_modules(void)
{
   unsigned int i;
   int status;

   boot.load_offset = 0;

   for (i = 0; i < boot.modules_nr; i++) {
      if (!boot.modules[i].is_loaded) {
         boot.modules[i].load_size = 0;
         break;
      }
      boot.load_offset += boot.modules[i].load_size;
   }

   load_sanity_check();

   for ( ; i < boot.modules_nr; i++) {
      status = load_module(i);
      if (status != ERR_SUCCESS) {
         return status;
      }
   }

   return ERR_SUCCESS;
}
