/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * mboot.c -- Multiboot loader
 *
 *   mboot [aSstRpeV] -c <FILEPATH> [KERNEL_OPTIONS]
 *
 *      OPTIONS
 *         -a             Do not pass extended attributes in the Multiboot info
 *                        memory map. This is necessary for kernels which do not
 *                        support Multiboot memory map extensions, and which
 *                        hardcode the memory map entry size.
 *         -c <FILEPATH>  Set the configuration file to FILEPATH.
 *         -S <1...4>     Set the default serial port (1=COM1, 2=COM2, 3=COM3,
 *                        4=COM4).
 *         -s <BAUDRATE>  Set the serial port speed to BAUDRATE (in bits per
 *                        second).
 *         -t <TITLE>     Set the bootloader banner title.
 *         -R <CMDLINE>   Set the command to be executed when <SHIFT+R> is
 *                        pressed. <CMDLINE> is only executed if the underlying
 *                        firmware library does not allow to return from main().
 *         -p <0...n>     Set the boot partition to boot from.
 *         -e             Exit on transient errors.
 *         -V             Enable Verbose mode.
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <boot_services.h>
#include "mboot.h"

#define DEFAULT_SERIAL_COM      1       /* Default serial port (COM1) */
#define DEFAULT_SERIAL_BAUDRATE 115200  /* Default serial baud rate */

boot_info_t boot;             /* Bootloader main info structure */

static char *kopts = NULL;

#define D(symbol, string) string,
static const char *error_str[] = {
   ERROR_TABLE
};
#undef D

/*-- clean ---------------------------------------------------------------------
 *
 *      Clean up everything so mboot can return properly.
 *----------------------------------------------------------------------------*/
static int clean(int status)
{
   if (status != ERR_SUCCESS) {
      Log(LOG_ERR, "Fatal error: %d (%s)\n", status, error_str[status]);
      if (boot.exit_on_errors && ((status == ERR_LOAD_ERROR)   ||
                                  (status == ERR_DEVICE_ERROR) ||
                                  (status == ERR_NOT_FOUND)    ||
                                  (status == ERR_NO_RESPONSE)  ||
                                  (status == ERR_TIMEOUT)      ||
                                  (status == ERR_TFTP_ERROR)   ||
                                  (status == ERR_END_OF_FILE)  ||
                                  (status == ERR_UNEXPECTED_EOF))) {
         if (!gui_exit()) {
            while (1);
         }
      } else {
         while (1);
      }
   }

   sys_free(kopts);
   sys_free(boot.cfgfile);
   unload_boot_modules();
   config_clear();

   return status;
}

/*-- mboot_init ----------------------------------------------------------------
 *
 *      Early bootloader initialization
 *
 * Parameters
 *      IN argc: number of command line arguments
 *      IN argv: pointer to arguments array
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int mboot_init(int argc, char **argv)
{
   int opt, serial_com, serial_speed, status;
   bool serial;

   serial = false;
   serial_com = DEFAULT_SERIAL_COM;
   serial_speed = DEFAULT_SERIAL_BAUDRATE;

   memset(&boot, 0, sizeof (boot_info_t));
   boot.bootif = true;

   if (argc < 1) {
      Log(LOG_ERR, "Command line is empty.\n");
      return ERR_SYNTAX;
   }

   optind = 1;

   do {
      opt = getopt(argc, argv, ":ac:R:p:S:s:t:Ve");
      switch (opt) {
         case -1:
            break;
         case 'a':
            boot.no_mem_attr = true;
            break;
         case 'c':
            boot.cfgfile = strdup(optarg);
            if (boot.cfgfile == NULL) {
               return ERR_OUT_OF_RESOURCES;
            }
            break;
         case 'p':
            if (!is_number(optarg)) {
               return ERR_SYNTAX;
            }
            boot.volid = atoi(optarg);
            break;
         case 'V':
            boot.verbose = true;
            break;
         case 'R':
            boot.recovery_cmd = strdup(optarg);
            if (boot.recovery_cmd == NULL) {
               return ERR_OUT_OF_RESOURCES;
            }
            break;
         case 'S':
            if (!is_number(optarg)) {
               return ERR_SYNTAX;
            }
            serial = true;
            serial_com = atoi(optarg);
            break;
         case 's':
            if (!is_number(optarg)) {
               return ERR_SYNTAX;
            }
            serial = true;
            serial_speed = atoi(optarg);
            break;
         case 't':
            gui_set_title(optarg);
            break;
         case 'e':
            boot.exit_on_errors = true;
            break;
         case 'd':
            /*
             * XXX: 'drive number/signature' (To be implemented)
             *      mboot currently loads modules from the boot media. This
             *      option will allow to specify an alternate boot drive to
             *      boot modules from.
             */
         case ':':
            /* Missing option argument */
         case '?':
            /* Missing option argument */
         default:
            return ERR_SYNTAX;
      }
   } while (opt != -1);

   if (serial) {
      serial_log_init(serial_com, serial_speed);
   }

   argc -= optind;
   argv += optind;

   if (argc > 0) {
      /*
       * Remaining arguments are treated as kernel options. This is required
       * for compatibility with the IPAPPEND pxelinux option which automatically
       * appends the BOOTIF=<MAC_ADDR> option to mboot's command line.
       */
      status = argv_to_str(argc, argv, &kopts);
      if (status != ERR_SUCCESS) {
         return status;
      }
   }

   snprintf(boot.name, MBOOT_ID_SIZE, "%s", MBOOT_ID_STR);

   return ERR_SUCCESS;
}

/*-- ipappend_2 ----------------------------------------------------------------
 *
 *      Detect the boot interface MAC address and force appending the
 *      'BOOTIF=xx-aa-bb-cc-dd-ee-ff' to the kernel command line, where xx is
 *      the Hardware Type Number of the boot interface (see RFC 1700), and
 *      aa:bb:cc:dd:ee:ff is its MAC address.
 *
 *      This avoids us to depend on the 'IPAPPEND' pxelinux option (which has no
 *      UEFI equivalent).
 *
 *      This function does nothing if the BOOTIF= option is already present on
 *      the kernel command line (for instance when IPAPPEND 2 is actually set in
 *      the pxelinux configuration file).
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int ipappend_2(void)
{
   const char *bootif, *opt;
   int status;

   if (boot.modules[0].options != NULL) {
      opt = strstr(boot.modules[0].options, "BOOTIF=");
      if (opt != NULL) {
         if (opt == boot.modules[0].options || isspace(*(--opt))) {
            return ERR_SUCCESS;
         }
      }
   }

   status = get_bootif_option(&bootif);
   if (status != ERR_SUCCESS) {
      Log(LOG_WARNING, "Network boot: MAC address not found.\n");
      Log(LOG_WARNING, "Add \'BOOTIF=01-aa-bb-cc-dd-ee-ff\' manually to the "
          "boot options (required if booting from gPXE).\n");
      Log(LOG_WARNING, "Replace aa-bb-cc-dd-ee-ff with the MAC address of the "
          "boot interface\n");

      return ERR_SUCCESS;
   }

   Log(LOG_DEBUG, "Network boot: %s\n", bootif);

   return append_kernel_options(bootif);
}

/*-- main ----------------------------------------------------------------------
 *
 *      Multiboot loader main function.
 *
 * Parameters
 *      IN argc: number of arguments on the command line
 *      IN argv: the command line
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *
 * Side Effects
 *      Upon success, should never return.
 *----------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
   handoff_t handoff;
   handoff_t *hdoff;
   int status;

   status = log_init();
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   status = mboot_init(argc, argv);
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   status = dump_firmware_info();
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   if (gui_init() != ERR_SUCCESS) {
      boot.headless = true;
   }

   status = parse_config(boot.cfgfile);
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   if (kopts != NULL) {
      status = append_kernel_options(kopts);
      sys_free(kopts);
      if (status != ERR_SUCCESS) {
         return clean(status);
      }
   }

   if (boot.bootif && is_network_boot()) {
      status = ipappend_2();
      if (status != ERR_SUCCESS) {
         return clean(status);
      }
   }

   if (!boot.headless) {
      status = gui_edit_kernel_options();
      if (status == ERR_ABORTED) {
         clean(ERR_SUCCESS);
         status = chainload_parent(boot.recovery_cmd);
         return (status == ERR_SUCCESS) ? ERR_ABORTED : status;
      } else if (status != ERR_SUCCESS) {
         return clean(status);
      }
   }

   status = load_boot_modules();
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   Log(LOG_DEBUG, "Initializing Multiboot standard...\n");

   status = multiboot_init();
   if (status != ERR_SUCCESS) {
      return clean(status);
   }

   Log(LOG_DEBUG, "Shutting down firmware services...\n");

   if (firmware_shutdown(&boot.mmap, &boot.mmap_count,
                         &boot.efi_info)               != ERR_SUCCESS
    || multiboot_register()                            != ERR_SUCCESS
    || compute_relocations(boot.mmap, boot.mmap_count) != ERR_SUCCESS
    || multiboot_set_runtime_pointers(&handoff)        != ERR_SUCCESS
    || install_trampoline(&handoff, &hdoff)            != ERR_SUCCESS) {
      /* Cannot return because Boot Services have been shutdown. */
      PANIC();
   }

   /* Process the relocations and jump to the kernel. */
   Log(LOG_INFO, "Relocating modules and starting up the kernel...\n");
   hdoff->kernel = boot.kernel.entry;
   hdoff->trampoline(hdoff);

   NOT_REACHED();

   return ERR_UNKNOWN;
}
