/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * mboot.h -- Multiboot Loader header file
 */

#ifndef MBOOT_H_
#define MBOOT_H_

#include <bootlib.h>
#include <multiboot.h>
#include <elf.h>
#include <vbe.h>
#include <e820.h>
#include <error.h>
#include <segment.h>
#include <efi_info.h>

#define X86_PAGE_SIZE          0x1000ULL   /* 4KB page size */
#define PAGE_ADDR(_addr_)      ((_addr_) & ~(X86_PAGE_SIZE - 1))
#define PAGE_ALIGN_UP(_addr_)  PAGE_ADDR((_addr_) + X86_PAGE_SIZE - 1)

/*
 * system.c
 */
int dump_firmware_info(void);
int firmware_shutdown(e820_range_t **mmap, size_t *count, efi_info_t *efi_info);

/*
 * reloc.c
 */
#define MAX_32_BIT_ADDR 0xffffffffULL
#define MAX_64_BIT_ADDR 0xffffffffffffffffULL

#define ALIGN_ANY       1
#define ALIGN_STR       1
#define ALIGN_PAGE      X86_PAGE_SIZE
#define ALIGN_FUNC      16
#define ALIGN_PTR       (sizeof (void *))

/*
 * XXX: Temporary work around:
 *      We have no way to determine the size of a function in standard C.
 *      Then we always return the maximum size of a function, considering that
 *      it ends at the end of the .text segment.
 *      A much better approach for getting the actual size of a function is
 *      implementing it in assembly.
 */
#define SIZEOF_C_FUNC(_func_)   (_etext - (char *)_func_)
#define SIZEOF_ASM_FUNC(_func_) ((char *)_func_ ## _end - (char *)_func_)

#define add_kernel_object(_src_, _size_, _dest_)                     \
   add_runtime_object('k', (_src_), (_size_), (_dest_), ALIGN_ANY)

#define add_sysinfo_object(_src_, _size_, _align_)                   \
   add_runtime_object('s', (_src_), (_size_), 0, (_align_))

#define add_module_object(_src_, _size_)                             \
   add_runtime_object('m', (_src_), (_size_), 0, ALIGN_PAGE)

#define add_safe_object(_src_, _size_, _align_)                      \
   add_runtime_object('t', (_src_), (_size_), 0, (_align_))

typedef uint64_t run_addr_t;  /* post-relocation object address */

typedef struct {
   run_addr_t dest;     /* Relocation destination */
   char *src;           /* Data source for memmove() or 0 for bzero() */
   uint64_t size;       /* Relocation length */
   size_t align;        /* Destination must be align on this size */
   char type;           /* Relocation type */
   char visited;        /* Visit counter for detecting circular dependencies */
} reloc_t;

typedef struct handoff_s handoff_t;

int add_runtime_object(char type, void *src, uint64_t size, run_addr_t dest,
                       size_t align);
int compute_relocations(e820_range_t *mmap, size_t count);
int runtime_addr(const void *addr, run_addr_t *runaddr);
int install_trampoline(handoff_t *handoff, handoff_t **runtime_handoff);

/*
 * alloc.c
 */
#define ALLOC_ANY       0
#define ALLOC_FIXED     1
#define ALLOC_FORCE     2

void alloc_sanity_check(void);
int alloc(uint64_t *addr, uint64_t size, size_t align, int option);

#define runtime_alloc_fixed(_addr_, _size_)                          \
   alloc((_addr_), (_size_), ALIGN_ANY, ALLOC_FIXED)

#define runtime_alloc(_addr_, _size_, _align_)                       \
   alloc((_addr_), (_size_), (_align_), ALLOC_ANY)

/*-- blacklist_runtime_mem ----------------------------------------------------
 *
 *      Reserve an explicit range of memory so it will not be allocated later
 *      for run-time memory.
 *
 * Parameters
 *      IN addr: memory range starting address
 *      IN size: memory range size
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static INLINE int blacklist_runtime_mem(uint64_t addr, uint64_t size)
{
   uint64_t page_start;

   page_start = PAGE_ADDR(addr);
   size = PAGE_ALIGN_UP(size + (addr - page_start));

   return alloc(&page_start, size, ALIGN_ANY, ALLOC_FORCE);
}

/*
 * trampoline.s
 */
#define TRAMPOLINE_STACK_SIZE 0x2000 /* 2 pages (much more than necessary) */
#define GDT_NENTRIES          3

EXTERN gdt_entry_t gdt[GDT_NENTRIES];

typedef void (CDECL *trampoline_t)(handoff_t *);

void CDECL relocate(handoff_t *handoff);
void relocate_end(void);
void trampoline(void);
void trampoline_end(void);

/*
 * elf.c
 */
int elf_check_headers(void *buffer, size_t buflen);
int elf_register(void *buffer, Elf_CommonAddr *entry, Elf_CommonShdr **shdr);

/*
 * mboot.c
 */
#define MBOOT_ID_STR    "MBOOT"
#define MBOOT_ID_SIZE   6
#define TITLE_MAX_LEN   128

#define NOT_REACHED()   for ( ; ; )
#define PANIC()         for ( ; ; )

#define BOOT_CS         MAKE_SEGMENT_SELECTOR(1, TI_GDT, RPL_SUPERVISOR)
#define BOOT_DS         MAKE_SEGMENT_SELECTOR(2, TI_GDT, RPL_SUPERVISOR)

typedef struct {
   Elf_CommonAddr entry;      /* Run-time entry point address */
   Elf_CommonShdr *shdr;      /* ELF section table address */
} kernel_t;

typedef struct {
   char *filename;            /* Module file name */
   char *options;             /* Module option string */
   void *addr;                /* Load address */
   size_t load_size;          /* Compressed module size (in bytes) */
   size_t size;               /* Decompressed module size (in bytes) */
   bool is_loaded;            /* True if the module has been entirely loaded */
} module_t;

typedef struct {
   char name[MBOOT_ID_SIZE];  /* Bootloader identification string */
   char title[TITLE_MAX_LEN]; /* Title string */
   char *cfgfile;             /* Configuration filename */
   int volid;                 /* Volume to load the kernel/modules from */
   kernel_t kernel;           /* Kernel information */
   unsigned int modules_nr;   /* Number of modules */
   module_t *modules;         /* Modules cmdlines, addresses and sizes */
   e820_range_t *mmap;        /* E820 Memory map */
   size_t mmap_count;         /* Number of entries in the memory map */
   framebuffer_t fb;          /* Framebuffer properties */
   efi_info_t efi_info;       /* EFI-specific information */
   uint64_t load_size;        /* Total size to load (in bytes) */
   uint64_t load_offset;      /* Current amount of loaded memory (in bytes) */
   char *recovery_cmd;        /* Command to be executed on <SHIFT+R> */
   bool verbose;              /* Verbose mode (true = on, false = off) */
   bool headless;             /* True if no video adapter is found */
   bool exit_on_errors;       /* Exit on transient errors */
   bool bootif;               /* Force BOOTIF= on the kernel command line */
   bool no_mem_attr;          /* Do not pass extended attr in the MBI mmap */
} boot_info_t;

EXTERN boot_info_t boot;

/*
 * Offsets in trampoline.inc must be updated if this structure is modified.
 */
#pragma pack (1)
struct handoff_s {
   run_addr_t stack;          /* Trampoline stack */
   run_addr_t relocs;         /* Table of relocations */
   run_addr_t relocate;       /* Relocation routine */
   run_addr_t mbi;            /* Multiboot Info structure address */
   run_addr_t kernel;         /* Kernel entry point */
   gdtr_t gdtr;               /* Boot GDTR */
   far_pointer_t ljump;       /* Far pointer to the final 32-bit code */
   trampoline_t trampoline;   /* Trampoline entry point */
};
#pragma pack()

int progress(void);

/*
 * load.c
 */
int get_load_size_hint(void);
int load_boot_modules(void);
void unload_boot_modules(void);

/*
 * config.c
 */
int append_kernel_options(const char *options);
int parse_config(const char *filename);
void config_clear(void);

/*
 * multiboot.c
 */
int check_multiboot_kernel(void *kbuf, size_t ksize);
size_t mb_mmap_desc_size(void);
int multiboot_set_runtime_pointers(handoff_t *handoff);
int multiboot_init(void);
int multiboot_register(void);

/*
 * gui.c
 */
int gui_init(void);
int gui_edit_kernel_options(void);
void gui_refresh(void);
void gui_set_title(const char *title);
int gui_resize(unsigned int width, unsigned int height,
               unsigned int depth);
int gui_text(void);
bool gui_exit(void);

#endif /* !MBOOT_H_ */
