/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * system.c -- Various system routines
 */

#include <bios.h>
#include <sm_bios.h>
#include <mps.h>
#include <cpu.h>
#include <boot_services.h>

#include "mboot.h"

#define LOW_SYSTEM_MEM  0x10000ULL   /* Considering this as system memory */

static void mmap_sanity_check(e820_range_t *mmap, size_t count)
{
   uint64_t max_base, max_limit, base, len, limit;
   bool error, overlap;
   const char *msg;
   size_t i;

   error = false;
   overlap = false;
   max_base = 0;
   max_limit = 0;

   for (i = 0; i < count; i++) {
      msg = NULL;
      base = E820_BASE(&mmap[i]);
      len = E820_LENGTH(&mmap[i]);
      limit = base + len - 1;

      if (!(i + 1 == count && base + len == 0) && (base + len < base)) {
         msg = "Memory map descriptor limit overflow\n";
      } else if (base < max_base) {
         msg = "Memory map is not sorted\n";
      }

      if (len > 0 && limit < max_limit) {
         overlap = true;
      }

      if (msg != NULL) {
         error = true;
         Log(LOG_ERR, "E820[%zu]: %llx - %llx type %u: %s\n",
             i, base, limit, mmap[i].type, msg);
      }

      max_base = base;
      max_limit = limit;
   }

   if (overlap || error) {
      for (i = 0; i < count; i++) {
         base = E820_BASE(&mmap[i]);
         limit = E820_BASE(&mmap[i]) + E820_LENGTH(&mmap[i]) - 1;
         Log(LOG_DEBUG, "E820[%zu]: %llx - %llx type %u\n",
             i, base, limit, mmap[i].type);
      }

      if (overlap) {
         Log(LOG_WARNING, "Memory map contains overlapping ranges\n");
      }

      if (error) {
         Log(LOG_ERR, "Memory map is corrupted.\n");
         while (1);
      }
   }
}

int dump_firmware_info(void)
{
   firmware_t firmware;
   int status;

   status = get_firmware_info(&firmware);
   if (status != ERR_SUCCESS) {
      Log(LOG_ERR, "Firmware detection failure.\n");
      return status;
   }

   switch (firmware.interface) {
      case FIRMWARE_INTERFACE_EFI:
         Log(LOG_DEBUG, "%s v%u.%u (%s, Rev.%u)\n",
             (firmware.version.efi.major > 1) ? "UEFI" : "EFI",
             (uint32_t)firmware.version.efi.major,
             (uint32_t)firmware.version.efi.minor,
             ((firmware.vendor != NULL) ? firmware.vendor : "Unknown vendor"),
             firmware.revision);
         break;
      case FIRMWARE_INTERFACE_COM32:
         Log(LOG_DEBUG, "COM32 v%u.%u (%s)\n",
             (uint32_t)firmware.version.com32.major,
             (uint32_t)firmware.version.com32.minor,
             ((firmware.vendor != NULL) ?
              firmware.vendor : "Unknown derivative"));
         break;
      default:
         Log(LOG_WARNING, "Unknown firmware\n");
   }

   sys_free(firmware.vendor);

   return ERR_SUCCESS;
}

/*-- reserve_sysmem ------------------------------------------------------------
 *
 *      Blacklist system memory so it will not be used later for run-time
 *      relocations.
 *
 * Parameters
 *      IN name: memory region description string
 *      IN addr: pointer to the memory region to be blacklisted
 *      IN size: memory region size
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static INLINE int reserve_sysmem(const char *name, void *addr, uint64_t size)
{
   Log(LOG_DEBUG, "%s found @ %p (%llu bytes)\n", name, addr, size);

   return blacklist_runtime_mem(PTR_TO_UINT64(addr), size);
}

/*-- scan_mps_memory -----------------------------------------------------------
 *
 *      Register the MPS memory.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int scan_mps_memory(void)
{
   MPFloatingPointer_t *fptr;
   MPConfigTable_t *table;
   int status;

   if (get_mps_table((void **)&fptr) != ERR_SUCCESS || fptr->length == 0) {
      /* No MPS found, do nothing. */
      return ERR_SUCCESS;
   }

   if (!is_valid_bios_table(fptr, fptr->length)) {
      return reserve_sysmem("MPS: invalid floating pointer", fptr,
                            sizeof (MPFloatingPointer_t));
   }

   status = reserve_sysmem("MPS: floating pointer", fptr,
                           fptr->length * MPS_PARAGRAPH_SIZE);
   if (status != ERR_SUCCESS) {
      return status;
   }

   table = UINT32_TO_PTR(fptr->physptr);
   if (table != NULL && table->length != 0) {
      if (!is_valid_bios_table(table, table->length)) {
         return reserve_sysmem("MPS: invalid table", table,
                               sizeof (MPConfigTable_t));
      }

      status = reserve_sysmem("MPS: table", table,
                              table->length + table->extsize);
      if (status != ERR_SUCCESS) {
         return status;
      }

      if (table->oemptr > 0 && table->oemsize > 0) {
         status = reserve_sysmem("MPS: OEM table", UINT32_TO_PTR(table->oemptr),
                                 table->oemsize);
         if (status != ERR_SUCCESS) {
            return status;
         }
      }
   }

   return ERR_SUCCESS;
}

/*-- scan_smbios_memory --------------------------------------------------------
 *
 *      Register the SMBIOS memory.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int scan_smbios_memory(void)
{
   SMBIOS_entry_point_t *eps;
   int status;

   if (get_smbios_table((void **)&eps) != ERR_SUCCESS || eps->length == 0) {
      /* No SMBIOS found, do nothing. */
      return ERR_SUCCESS;
   }

   if (!is_valid_bios_table(eps, eps->length)) {
      return reserve_sysmem("SMBIOS: Invalid floating pointer", eps,
                            sizeof (MPFloatingPointer_t));
   }

   status = reserve_sysmem("SMBIOS: floating pointer", eps, eps->length);
   if (status != ERR_SUCCESS) {
      return status;
   }

   return reserve_sysmem("SMBIOS: table",
                         UINT32_TO_PTR(eps->SMBIOSStructTableAddress),
                         eps->SMBIOSStructTableLength);
}

/*-- scan_memory_map -----------------------------------------------------------
 *
 *      Blacklist memory that the memory map does not report as available:
 *        - memory above the last address reported in the memory map
 *        - undefined memory (holes in the memory map)
 *        - memory that is not reported as 'available' in the memory map
 *          (bootloader's memory is considered as 'available')
 *
 * Parameters
 *      IN mmap:  pointer to the system memory map
 *      IN count: number of entries in the memory map
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int scan_memory_map(e820_range_t *mmap, size_t count)
{
   uint64_t addr;
   int status;
   size_t i;

   Log(LOG_DEBUG, "Scanning system memory (%zu entries)...\n", count);

   mmap_sanity_check(mmap, count);

   addr = E820_BASE(&mmap[count - 1]) + E820_LENGTH(&mmap[count - 1]) - 1;
   status = blacklist_runtime_mem(addr + 1, MAX_64_BIT_ADDR - addr);
   if (status != ERR_SUCCESS) {
      return status;
   }

   addr = 0;
   for (i = 0; i < count; i++) {
      if (E820_LENGTH(mmap) == 0) {
         continue;
      }

      if (E820_BASE(mmap) - addr > 0) {
         status = blacklist_runtime_mem(addr, E820_BASE(mmap) - addr);
         if (status != ERR_SUCCESS) {
            return status;
         }
      }

      if (mmap->type != E820_TYPE_AVAILABLE && mmap->type != E820_TYPE_BOOTLOADER) {
         status = blacklist_runtime_mem(E820_BASE(mmap), E820_LENGTH(mmap));
         if (status != ERR_SUCCESS) {
            return status;
         }
      }

      addr = E820_BASE(mmap) + E820_LENGTH(mmap);
      mmap++;
   }

   return ERR_SUCCESS;
}

/*-- blacklist_system_memory ---------------------------------------------------
 *
 *      List all the memory ranges that may not be used by the bootloader.
 *
 * Parameters
 *      IN mmap:  pointer to the system memory map
 *      IN count: number of entries in the memory map
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int blacklist_system_memory(e820_range_t *mmap, size_t count)
{
   int status;

   status = blacklist_runtime_mem(0, LOW_SYSTEM_MEM);
   if (status != ERR_SUCCESS) {
      return status;
   }
   status = blacklist_runtime_mem(BIOS_ROM_START, BIOS_ROM_SIZE);
   if (status != ERR_SUCCESS) {
      return status;
   }
   status = blacklist_runtime_mem(0, 0x100000);
   if (status != ERR_SUCCESS) {
      return status;
   }
   status = scan_smbios_memory();
   if (status != ERR_SUCCESS) {
      return status;
   }
   status = scan_mps_memory();
   if (status != ERR_SUCCESS) {
      return status;
   }
   status = scan_memory_map(mmap, count);
   if (status != ERR_SUCCESS) {
      return status;
   }

   return ERR_SUCCESS;
}

/*-- firmware_shutdown ---------------------------------------------------------
 *
 *     Shutdown the boot services:
 *       - Get the run-time E820 memory map (and request some extra memory for
 *         converting it later to the -bigger- Multiboot format).
 *
 *       - Record some EFI-specific information if possible.
 *
 *       - Claim that we no longer need the firmware boot services.
 *
 *       - Disable hardware interrupts. Since firmware services have been shut
 *         down, it is no longer necessary to run firmware interrupt handlers.
 *         After this function is called, it is safe to clobber the IDT and GDT.
 *
 * Parameters
 *      OUT mmap:     pointer to the freshly allocated memory map
 *      OUT count:    number of entries in the memory map
 *      OUT efi_info: EFI information if available
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *
 * Side Effects
 *      No call to the boot services may be done after a call to this function.
 *----------------------------------------------------------------------------*/
int firmware_shutdown(e820_range_t **mmap, size_t *count, efi_info_t *efi_info)
{
   size_t desc_extra_mem;
   int status;

   if (mb_mmap_desc_size() > sizeof (e820_range_t)) {
      desc_extra_mem = mb_mmap_desc_size() - sizeof (e820_range_t);
   } else {
      desc_extra_mem = 0;
   }

   status = exit_boot_services(desc_extra_mem, mmap, count, efi_info);
   if (status != ERR_SUCCESS) {
      Log(LOG_ERR, "Failed to shutdown the boot services.\n");
      return status;
   }

   CLI();

   Log(LOG_DEBUG, "Scanning system tables...");

   e820_mmap_merge(*mmap, count);

   status = blacklist_system_memory(*mmap, *count);
   if (status != ERR_SUCCESS) {
      Log(LOG_ERR, "Error scanning system memory.\n");
      return status;
   }

   return ERR_SUCCESS;
}
