;;******************************************************************************
;; Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
;;
;; This file is part of the esxboot project.
;;
;; esxboot is free software ; you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation version 2 and no later version.
;;
;; esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
;; A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
;; details.
;;
;; You should have received a copy of the GNU General Public License along with
;; this program         ; if not, write to the Free Software Foundation, Inc.,
;; 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
;;******************************************************************************

;; trampoline.s
;;
;;      All the code in this module will be relocated in order to execute from
;;      safe memory. That is to ensure that it will never get overwritten when
;;      processing the relocations.
;;
;;      - THIS CODE MUST BE POSITION-INDEPENDENT
;;      - THIS CODE MUST BE RELOCATED BELOW 4-GB
;;      - THIS CODE ASSUMES THAT HARDWARE INTERRUPTS ARE DISABLED

DEFAULT REL

%include "trampoline.inc"

SECTION .text

;;-- relocate ------------------------------------------------------------------
;;
;;      This is the function that actually processes the relocations and then
;;      jumps to the trampoline.
;;
;; Prototype
;;      void relocate(handoff_t *handoff);
;;
;; Side effects:
;;      - Setup a new GDT
;;------------------------------------------------------------------------------
GLOBAL relocate
relocate:

%if IS_64_BIT
[BITS 64]
      mov     r12, rdi                    ; R12 = handoff (R12 is callee-safe)
      mov     rsp, [r12 + OFFSET_STACK]   ; RSP = handoff->stack (new stack)

      sub     rsp, 4 * SIZEOF_PTR         ; Home allocation

      mov     rdi, [r12 + OFFSET_RELOCS]  ; RDI = handoff->reloc_table
      call    [r12 + OFFSET_RELOCATE]     ; handoff->do_reloc(RDI)

      mov     rbx, [r12 + OFFSET_MBI]     ; RBX = handoff->mbi
      mov     rbp, [r12 + OFFSET_KERNEL]  ; RBP = handoff->kernel

      lgdt    [r12 + OFFSET_GDTR]         ; Load 32-bit GDT
      jmp far DWORD [r12 + OFFSET_LJUMP]  ; Reload CS and serialize the CPU
                                          ; We use DWORD, because jmp far QWORD
                                          ; with absolute indirect addressing is
                                          ; not supported on AMD platforms.
%else
[BITS 32]
      mov     esi, [esp + SIZEOF_PTR]     ; ESI = handoff
      mov     esp, [esi + OFFSET_STACK]   ; ESP = handoff->stack (new stack)

      push    esi                         ; Save ESI on the new stack
      push    DWORD [esi + OFFSET_RELOCS] ; ARG1 = handoff->reloc_table
      call    [esi + OFFSET_RELOCATE]     ; handoff->do_reloc(ARG1)
      add     esp, 4                      ; Adjust the stack (pop ARG1)
      pop     esi                         ; Restore ESI from the stack

      mov     ebx, [esi + OFFSET_MBI]     ; EBX = handoff->mbi
      mov     ebp, [esi + OFFSET_KERNEL]  ; EBP = handoff->kernel

      lgdt    [esi + OFFSET_GDTR]         ; Load 32-bit GDT
      jmp far DWORD [esi + OFFSET_LJUMP]  ; Reload CS and serialize the CPU
%endif

GLOBAL relocate_end
relocate_end:

;;-- trampoline ----------------------------------------------------------------
;;
;;      Setup a Multiboot compliant environment and jumps to the kernel.
;;
;;      Here we are either in compatibility mode, or in protected mode.
;;      New GDT is up, and descibes a 4-Gb flat 32-bit memory.
;;
;; Parameters
;;      EBX = Multiboot Info structure address
;;      EBP = kernel entry point
;;
;; Side effects
;;      - Disable paging
;;      - Disable Long Mode
;;------------------------------------------------------------------------------
GLOBAL trampoline
trampoline:
[BITS 32]
      mov      eax, cr0                   ; EAX = CR0
      btr      eax, CR0_PAGING_BIT        ; Clear (disable) paging bit
      mov      cr0, eax                   ; Disable paging and clear EFER.LMA

%if IS_64_BIT
      mov      ecx, MSR_EFER              ; ECX selects MSR EFER
      rdmsr                               ; EDX:EAX = MSR EFER
      btr      eax, MSR_EFER_LME_BIT      ; Clear (disable) Long Mode bit
      wrmsr                               ; MSR EFER = EDX:EAX
%endif

;;
;; Here we are in protected mode, paging disabled.
;;
      mov      ax, BOOT_DS                ; For reloading data segments
      mov      ds, ax                     ; DS = BOOT_DS
      mov      es, ax                     ; ES = BOOT_DS
      mov      fs, ax                     ; FS = BOOT_DS
      mov      gs, ax                     ; GS = BOOT_DS
      mov      ss, ax                     ; SS = BOOT_DS

      mov      eax, MULTIBOOT_MAGIC       ; EAX = Multiboot Magic
                                          ; MBI address is already in EBX
      call     ebp                        ; handoff->kernel()

;;
;; Not supposed to be reached.
;;
.not_reached:                             ; Kernel is not supposed to return!
      hlt
      jmp     .not_reached

GLOBAL trampoline_end
trampoline_end:

;;-- DATA ----------------------------------------------------------------------
SECTION .data

GLOBAL gdt
gdt:
      dq GDT_NULL_DESC
      dq GDT_4GB_FLAT_CODE32_DESC
      dq GDT_4GB_FLAT_DATA32_DESC
