/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * safeboot.c -- Safeboot module
 *
 *   safeboot [-rsSV] [-m <FILEPATH>]
 *
 *      OPTIONS
 *         -m <FILEPATH>  Next loader path
 *         -r             Force rollback (not used on UEFI systems)
 *         -S <1...4>     Set the default serial port (1=COM1, 2=COM2, 3=COM3,
 *                        4=COM4).
 *         -s <BAUDRATE>  Set the serial port speed to BAUDRATE (in bits per
 *                        second).
 *         -V             Verbose mode
 */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <boot_services.h>
#include "safeboot.h"

#define DEFAULT_SERIAL_COM      1       /* Default serial port (COM1) */
#define DEFAULT_SERIAL_BAUDRATE 115200  /* Default serial baud rate */

#define DEFAULT_PROG_NAME       "safeboot.c32"

safeboot_env_t safeboot;

static bool shift_r = false;

/*-- safeboot_init -------------------------------------------------------------
 *
 *      Parse safeboot command line options.
 *
 * Parameters
 *      IN argc: number of command line arguments
 *      IN argv: pointer to the command line arguments array
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
static int safeboot_init(int argc, char **argv)
{
   int opt;

   memset(&safeboot, 0, sizeof (safeboot_env_t));

   if (argc == 0 || argv == NULL || argv[0] == NULL) {
      return ERR_INVALID_PARAMETER;
   }

   safeboot.self_path = strdup((*argv[0] != '\0') ? argv[0] :
                               DEFAULT_PROG_NAME);
   if (safeboot.self_path == NULL) {
      return ERR_OUT_OF_RESOURCES;
   }

   safeboot.serial_com = DEFAULT_SERIAL_COM;
   safeboot.serial_speed = DEFAULT_SERIAL_BAUDRATE;

   if (argc > 1) {
      optind = 1;
      do {
         opt = getopt(argc, argv, ":m:rs:S:V");
         switch (opt) {
            case -1:
               break;
            case 'm':
               safeboot.next_loader = strdup(optarg);
               if (safeboot.next_loader == NULL) {
                  return ERR_OUT_OF_RESOURCES;
               }
               break;
            case 'r':
               shift_r = true;
               break;
            case 'S':
               if (!is_number(optarg)) {
                  return ERR_SYNTAX;
               }
               safeboot.serial = true;
               safeboot.serial_com = atoi(optarg);
               break;
            case 's':
               if (!is_number(optarg)) {
                  return ERR_SYNTAX;
               }
               safeboot.serial = true;
               safeboot.serial_speed = atoi(optarg);
               break;
            case 'V':
               safeboot.verbose = true;
               break;
            case ':':
            case '?':
            default:
               return ERR_SYNTAX;
         }
      } while (opt != -1);
   }

   return ERR_SUCCESS;
}

/*-- main ----------------------------------------------------------------------
 *
 *      Safeboot module main function.
 *
 * Parameters
 *      IN argc: number of command line arguments
 *      IN argv: pointer to the command line arguments array
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
   bootbank_t *bank;
   key_code_t key;
   int retval, status;

   status = log_init();
   if (status != ERR_SUCCESS) {
      return status;
   }

   status = safeboot_init(argc, argv);
   if (status != ERR_SUCCESS) {
      return status;
   }

   do {
      if (safeboot.serial) {
         serial_log_init(safeboot.serial_com, safeboot.serial_speed);
      }

      /*
       * (Re-)initialize the video display if any. gui_init() return value is
       * ignored, letting headless platforms (with no video adpater) continue
       * booting with graphics disabled.
       */
      gui_init();

      status = get_boot_bank(shift_r, &bank);
      if (status != ERR_SUCCESS) {
         Log(LOG_ERR, "No hypervisor found.\n");
         break;
      }

      retval = ERR_SUCCESS;
      status = chainload(bank, &retval);
      bank_clean();

      if (status != ERR_SUCCESS) {
         Log(LOG_ERR, "Failed to chainload BANK%d.\n", bank->volid);
         break;
      }

      if (retval == ERR_SUCCESS) {
         /*
          * We can only be here if mboot has returned successfully. This can
          * only happen when <SHIFT+R> is pressed from mboot to interrupt the
          * loading process.
          */
         shift_r = true;

         retval = log_init();
      }
   } while (retval == ERR_SUCCESS);

   sys_free(safeboot.next_loader);
   sys_free(safeboot.self_path);

   kbd_waitkey(&key);

   return status;
}
