/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * vmfat.c -- VMware FAT support
 */

#include <string.h>
#include <stdio.h>
#include <boot_services.h>
#include "safeboot.h"

#define VMWARE_FAT_MAGIC            "VMWARE FAT16    "
#define VMWARE_FAT_MAGIC_LEN        16
#define VMWARE_FAT_BLOCK_SIZE       512

/*-- vmfat_uuid_to_str ---------------------------------------------------------
 *
 *      Convert a VMware FAT UUID to a human readable string where each byte in
 *      the UUID is seen as an unsigned char and represented by a 0-prefixed,
 *      2-characters, lower case hexadecimal string. The output string is
 *      allocated with sys_malloc() and can be freed with sys_free().
 *
 * Parameters
 *      IN uuid: pointer to the VMware FAT UUID
 *      IN str:  pointer to the freshly allocated output string
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int vmfat_uuid_to_str(const unsigned char *uuid, char **str)
{
   int i, len, status;
   char *s;

   ASSERT(uuid != NULL);
   ASSERT(str != NULL);

   status = str_alloc(2 * VMWARE_FAT_UUID_LEN, &s);
   if (status != ERR_SUCCESS) {
      return status;
   }

   *str = s;

   for (i = 0; i < VMWARE_FAT_UUID_LEN; i++) {
      len = snprintf(s, 3, "%02x", (unsigned int)uuid[i]);
      if (len != 2) {
         sys_free(s);
         return ERR_INVALID_PARAMETER;
      }
      s += 2;
   }

   return ERR_SUCCESS;
}

/*-- vmfat_get_uuid ------------------------------------------------------------
 *
 *      Read the VMware FAT UUID from the specified partition.
 *
 * Parameters
 *      IN volid:  MBR/GPT partition number
 *      IN buffer: pointer to the output buffer
 *      IN buflen: output buffer size, in bytes
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int vmfat_get_uuid(int volid, void *buffer, size_t buflen)
{
   uint8_t block[VMWARE_FAT_BLOCK_SIZE];
   uint8_t *uuid;
   int status;

   if (buffer == NULL || buflen < VMWARE_FAT_UUID_LEN) {
      return ERR_INVALID_PARAMETER;
   }

   status = volume_read(volid, block, VMWARE_FAT_BLOCK_SIZE,
                        VMWARE_FAT_BLOCK_SIZE);
   if (status != ERR_SUCCESS) {
      return status;
   }

   if (memcmp(VMWARE_FAT_MAGIC, block, VMWARE_FAT_MAGIC_LEN)) {
      return ERR_NOT_FOUND;
   }

   uuid = &block[VMWARE_FAT_MAGIC_LEN];
   memcpy(buffer, uuid, VMWARE_FAT_UUID_LEN);

   return ERR_SUCCESS;
}
