/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * console.c -- EFI console management
 */

#include <stdio.h>
#include <stdarg.h>
#include "efi_private.h"

#define EFI_MESSAGE_BUFLEN   128

#define DEFAULT_COM1         0x3f8
#define DEFAULT_COM2         0x2f8
#define DEFAULT_COM3         0x3e8
#define DEFAULT_COM4         0x2e8

void (*efi_log_callback)(int level, const char *, ...) = NULL;

static EFI_GUID ConsoleControlProto = EFI_CONSOLE_CONTROL_PROTOCOL_GUID;

/*-- efi_set_graphic_mode ------------------------------------------------------
 *
 *      Enable graphic mode.
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS efi_set_graphic_mode(void)
{
   EFI_CONSOLE_CONTROL_PROTOCOL *console;
   EFI_CONSOLE_CONTROL_SCREEN_MODE current_mode;
   EFI_STATUS Status;

   Status = LocateProtocol(&ConsoleControlProto, (void **)&console);
   if (EFI_ERROR(Status)) {
      /*
       * Most of UEFI 2.0 firmwares do not have a ConsoleControlProtocol, and
       * the hardware should already be in a linear graphic mode.
       * (EDK II has the GraphicsConsole and ConSplitter modules that assume
       * this in this case).
       * Therefore, we return success here and hope for the best.
       */
      return EFI_SUCCESS;
   }

   console->GetMode(console, &current_mode, NULL, NULL);
   if (current_mode != EfiConsoleControlScreenGraphics) {
      console->SetMode(console, EfiConsoleControlScreenGraphics);
   }

   return EFI_SUCCESS;
}

/*-- efi_set_text_mode ---------------------------------------------------------
 *
 *      Enable text mode.
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS efi_set_text_mode(void)
{
   EFI_CONSOLE_CONTROL_PROTOCOL *console;
   EFI_CONSOLE_CONTROL_SCREEN_MODE current_mode;
   EFI_STATUS Status;

   Status = LocateProtocol(&ConsoleControlProto, (void **)&console);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   console->GetMode(console, &current_mode, NULL, NULL);
   if (current_mode != EfiConsoleControlScreenGraphics) {
      console->SetMode(console, EfiConsoleControlScreenText);
   }

   return EFI_SUCCESS;
}

/*-- efi_print -----------------------------------------------------------------
 *
 *      Print an UEFI string on the default 'ConOut' device.
 *
 * Parameters
 *      IN Str: pointer to the UCS-2 string to print
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS efi_print(CHAR16 *Str)
{
   if (Str == NULL) {
      return EFI_INVALID_PARAMETER;
   }

   if (st == NULL || st->ConOut == NULL || st->ConOut->OutputString == NULL) {
      return EFI_UNSUPPORTED;
   }

   return st->ConOut->OutputString(st->ConOut, Str);
}

/*-- firmware_print ------------------------------------------------------------
 *
 *      Generic wrapper for efi_print() to print a standard ASCII string. The
 *      string is first converted to the UCS-2 format, and '\n' are replaced
 *      the "\n\r" sequence.
 *
 * Parameters
 *      IN str: pointer to the ASCII string to print
 *
 * Results
 *      A generic error code.
 *----------------------------------------------------------------------------*/
int firmware_print(const char *str)
{
   CHAR16 Buffer[EFI_MESSAGE_BUFLEN];
   CHAR16 *Str;
   char *buf;
   EFI_STATUS Status;
   int i, j;

   Str = &Buffer[0];
   buf = (char *)Str;
   i = 0;

   while (str[i] != '\0'){
      for (j = 0; j < (EFI_MESSAGE_BUFLEN - 2); i++) {
         buf[j++] = str[i];

         if (str[i] == '\n') {
            buf[j++] = '\r';
         } else if (str[i] == '\0') {
            break;
         }
      }

      if (j > 0 && buf[j - 1] != '\0') {
         buf[j] = '\0';
      }

      ascii_to_ucs2(buf, &Str);

      Status = efi_print(Str);
      if (EFI_ERROR(Status)) {
         return error_efi_to_generic(Status);
      }
   }

   return error_efi_to_generic(EFI_SUCCESS);
}

/*-- efi_log_default -----------------------------------------------------------
 *
 *      Default routine to be called in order to log a message from the UEFI
 *      layer.
 *
 * Parameters
 *      IN level: syslog message priority
 *      IN msg:   the message to be logged
 *      IN ...:   arguments list for the message format
 *----------------------------------------------------------------------------*/
static void efi_log_default(UNUSED_PARAM(int level), const char *msg, ...)
{
   char buffer[EFI_MESSAGE_BUFLEN];
   va_list ap;

   va_start(ap, msg);
   vsnprintf(buffer, EFI_MESSAGE_BUFLEN, msg, ap);
   va_end(ap);

   firmware_print(buffer);
}

/*-- efi_assert ----------------------------------------------------------------
 *
 *      Assert routine (debugging purposes)
 *
 * Parameters
 *      IN msg:   pointer to a message to be displayed
 *      IN ...:   arguments list for the message format
 *----------------------------------------------------------------------------*/
#ifdef DEBUG
void efi_assert(const char *msg, ...)
{
   char buffer[EFI_MESSAGE_BUFLEN];
   va_list ap;

   va_start(ap, msg);
   vsnprintf(buffer, EFI_MESSAGE_BUFLEN, msg, ap);
   va_end(ap);

   efi_log(LOG_EMERG, "%s", buffer);
}
#endif

/*-- set_firmware_log_callback -------------------------------------------------
 *
 *      Set the routine to be called in order to log a message from the UEFI
 *      layer.
 *
 * Parameters
 *      IN callback: pointer to the logging routine, or NULL to disable UEFI
 *                   logging
 *----------------------------------------------------------------------------*/
void set_firmware_log_callback(void (*callback)(int, const char *, ...))
{
   efi_log_callback = callback;
}

/*-- set_display_mode ---------------------------------------------------------
 *
 *      Set the display to either firmware native text mode, or VBE graphic
 *      mode.
 *
 * Parameters
 *      IN mode: the desired mode
 *
 * Results
 *      A generic status code.
 *----------------------------------------------------------------------------*/
int set_display_mode(display_mode_t mode)
{
   EFI_STATUS Status;

   switch (mode) {
      case DISPLAY_MODE_NATIVE_TEXT:
         Status = efi_set_text_mode();
         if (!EFI_ERROR(Status)) {
            set_firmware_log_callback(efi_log_default);
         }
         break;
      case DISPLAY_MODE_VBE:
         Status = efi_set_graphic_mode();
         if (!EFI_ERROR(Status) && efi_log_callback == efi_log_default) {
            set_firmware_log_callback(NULL);
         }
         break;
      default:
         Status = EFI_INVALID_PARAMETER;
   }

   return error_efi_to_generic(Status);
}

/*-- get_serial_port -----------------------------------------------------------
 *
 *      Get the I/O base address of a COM serial port. On UEFI platforms, we
 *      always assume that COM ports have fixed I/O base addresses.
 *
 * Parameters
 *      IN  com:  1=COM1, 2=COM2, 3=COM3, 4=COM4
 *      OUT port: the serial port I/O base address
 *
 * Results
 *      A generic error status.
 *----------------------------------------------------------------------------*/
int get_serial_port(int com, uint16_t *port)
{
   const uint16_t default_com_port[4] = { DEFAULT_COM1, DEFAULT_COM2,
                                          DEFAULT_COM3, DEFAULT_COM4 };

   if (com < 1 || com > 4) {
      return error_efi_to_generic(EFI_INVALID_PARAMETER);
   }

   *port = default_com_port[com - 1];

   return error_efi_to_generic(EFI_SUCCESS);
}
