/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * init.c -- EFI Firmware init/cleanup functions
 */

#include <string.h>
#include <ctype.h>

#include "efi_private.h"

EXTERN int main(int argc, char **argv);

EFI_BOOT_SERVICES *bs = NULL;
EFI_SYSTEM_TABLE *st = NULL;
EFI_HANDLE ImageHandle;

/*-- get_firmware_info ---------------------------------------------------------
 *
 *      Return the EFI interface information.
 *
 * Parameters
 *      IN firmware: pointer to the firmware info structure to be filled up
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 * ---------------------------------------------------------------------------*/
int get_firmware_info(firmware_t *firmware)
{
   EFI_STATUS Status;
   char *vendor;

   EFI_ASSERT(st != NULL);

   vendor = NULL;
   if (st->FirmwareVendor != NULL) {
      Status = ucs2_to_ascii(st->FirmwareVendor, &vendor);
      if (EFI_ERROR(Status)) {
         return error_efi_to_generic(Status);
      }
   }

   memset(firmware, 0, sizeof (firmware_t));
   firmware->interface = FIRMWARE_INTERFACE_EFI;
   firmware->version.efi.major = st->Hdr.Revision >> 16;
   firmware->version.efi.minor = st->Hdr.Revision & 0xffff;
   firmware->vendor = vendor;
   firmware->revision = st->FirmwareRevision;

   return error_efi_to_generic(EFI_SUCCESS);
}

/*-- efi_set_watchdog_timer ----------------------------------------------------
 *
 *      Reset the UEFI watchdog timer. Setting the Timeout parameter to zero
 *      disables the watchdog timer. The timer can be re-enabled by resetting it
 *      with a non-zero Timeout value.
 *
 *      Note: efi_set_watchdog_timer() returns successfully if no watchdog timer
 *            is supported on the platform.
 *
 * Parameters
 *      IN Timeout: expiration time, in seconds
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
EFI_STATUS efi_set_watchdog_timer(UINTN Timeout)
{
   EFI_STATUS Status;

   EFI_ASSERT(bs != NULL);
   EFI_ASSERT_FIRMWARE(bs->SetWatchdogTimer != NULL);

   Status = bs->SetWatchdogTimer(Timeout, 0, 0, NULL);

   if (Status == EFI_UNSUPPORTED) {
      return ERR_SUCCESS;
   }

   if (EFI_ERROR(Status)) {
      efi_log(LOG_WARNING, "Could not %s the UEFI watchdog timer.\n",
              (Timeout == WATCHDOG_DISABLE) ? "disable" : "reset");
   }

   return Status;
}

/*-- exit_boot_services --------------------------------------------------------
 *
 *      Exit UEFI boot services.
 *
 * Parameters
 *      IN  desc_extra_mem: amount of extra memory (in bytes) needed for each
 *                          memory map descriptor
 *      OUT mmap:           pointer to the E820 system memory map
 *      OUT count:          number of entry in the memory map
 *      OUT efi_info:       EFI information
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *
 * Side Effects
 *      Shutdown the boot services and invalidate the 'bs' global pointer.
 *      EFI Boot Services are no longer available after a call to this function.
 *----------------------------------------------------------------------------*/
int exit_boot_services(size_t desc_extra_mem, e820_range_t **mmap,
                       size_t *count, efi_info_t *efi_info)
{
   EFI_STATUS Status;
   int error;

   EFI_ASSERT(st != NULL);
   EFI_ASSERT(bs != NULL);
   EFI_ASSERT_FIRMWARE(bs->ExitBootServices != NULL);

   /*
    * UEFI Specification v2.3 (6.4. "Image Services", ExitBootServices()) says:
    *
    * "An EFI OS loader must ensure that it has the system's current memory map
    *  at the time it calls ExitBootServices(). This is done by passing in the
    *  current memory map's MapKey value as returned by GetMemoryMap().
    *  Care must be taken to ensure that the memory map does not change between
    *  these two calls. It is suggested that GetMemoryMap() be called
    *  immediately before calling ExitBootServices()."
    */
again:
   error = get_memory_map(desc_extra_mem, mmap, count, efi_info);
   if (error != ERR_SUCCESS) {
      return error;
   }

   Status = bs->ExitBootServices(ImageHandle, MapKey);
   if (Status == EFI_INVALID_PARAMETER) {
      sys_free(efi_info->mmap);
      goto again;
   }
   if (!EFI_ERROR(Status)) {
      bs = NULL;

      /*
       * UEFI Specification v2.3 (6.4. "Image Services") says:
       *
       * "On ExitBootServices() success, several fields of the EFI System Table
       *  should be set to NULL. These include ConsoleInHandle, ConIn,
       *  ConsoleOutHandle, ConOut, StandardErrorHandle, StdErr, and
       *  BootServicesTable. In addition, since fields of the EFI System Table
       *  are being modified, the 32-bit CRC for the EFI System Table must be
       *  recomputed."
       */
      st->ConsoleInHandle = NULL;
      st->ConIn = NULL;
      st->ConsoleOutHandle = NULL;
      st->ConOut = NULL;
      st->StandardErrorHandle = NULL;
      st->StdErr = NULL;
      st->BootServices = NULL;
      st->Hdr.CRC32 = 0;
      st->Hdr.CRC32 = efi_crc(&st->Hdr, st->Hdr.HeaderSize);

      efi_info->systab = PTR_TO_UINT64(st);
      efi_info->systab_size = st->Hdr.HeaderSize;
      efi_info->valid = true;
   }

   return error_efi_to_generic(Status);
}

/*-- efi_init ------------------------------------------------------------------
 *
 *      Initialize the firmware interface:
 *        1. Setup the ImageHandle, st and bs global variables so they point
 *           respectively to the current image handle, the UEFI system table,
 *           and the UEFI Boot Services dispatch structure.
 *        2. Reset the UEFI watchdog timer to 5 minutes.
 *        3. Initialize the memory allocator.
 *        4. Initialize the VBE emulation.
 *
 * Parameters
 *      IN Handle:      handle to the application image
 *      IN SystemTable: pointer to the EFI system table
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS efi_init(EFI_HANDLE Handle, EFI_SYSTEM_TABLE *SystemTable)
{
   EFI_LOADED_IMAGE *Image;
   EFI_STATUS Status;

   if (Handle == NULL || SystemTable == NULL ||
       SystemTable->BootServices == NULL) {
      return EFI_INVALID_PARAMETER;
   }

   ImageHandle = Handle;
   st = SystemTable;
   bs = st->BootServices;

   efi_set_watchdog_timer(WATCHDOG_DEFAULT_TIMEOUT);

   Status = image_get_info(ImageHandle, &Image);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   mem_init(Image->ImageDataType);

   return EFI_SUCCESS;
}

/*-- chainload_parent ----------------------------------------------------------
 *
 *      Transfer execution back to the parent process.
 *
 * Parameters
 *      IN cmdline: unused on UEFI systems
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int chainload_parent(UNUSED_PARAM(const char *cmdline))
{
   EFI_STATUS Status;

   EFI_ASSERT(bs != NULL);
   EFI_ASSERT_FIRMWARE(bs->Exit != NULL);

   Status = bs->Exit(ImageHandle, EFI_SUCCESS, 0, NULL);

   return error_efi_to_generic(Status);
}

/*-- efi_main ------------------------------------------------------------------
 *
 *      This is the EFI-specific application entry point. this function
 *      initializes the firmware interface and retrieves the command line
 *      arguments.
 *
 * Parameters
 *      IN Handle:      handle to the parent image
 *      IN SystemTable: pointer to the EFI system table
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
EFI_STATUS EFIAPI efi_main(EFI_HANDLE Handle, EFI_SYSTEM_TABLE *SystemTable)
{
   EFI_LOADED_IMAGE *Image;
   CHAR16 *CommandLine;
   CHAR16 *FilePath;
   char *cmdline;
   char **argv;
   int argc;
   int retval;
   EFI_STATUS Status;

   Status = efi_init(Handle, SystemTable);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   Status = image_get_info(ImageHandle, &Image);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   argc = 0;
   argv = NULL;
   cmdline = NULL;
   FilePath = NULL;

   if (Image->LoadOptions != NULL && Image->LoadOptionsSize > 0) {
      CommandLine = Image->LoadOptions;
      Status = EFI_SUCCESS;
   } else {
      Status = devpath_get_filepath(Image->FilePath, &FilePath);
      CommandLine = FilePath;
   }

   if (!EFI_ERROR(Status)) {
      Status = ucs2_to_ascii(CommandLine, &cmdline);

      if (!EFI_ERROR(Status)) {
         retval = str_to_argv(cmdline, &argc, &argv);
         Status = error_generic_to_efi(retval);
      }
   }

   if (!EFI_ERROR(Status)) {
      retval = main(argc, argv);
      Status = error_generic_to_efi(retval);
   }

   sys_free(cmdline);
   sys_free(argv);
   sys_free(FilePath);
   efi_clean_vbe();

   return Status;
}
