/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * net.c -- Network-related UEFI functions
 */

#include <stdio.h>
#include "efi_private.h"
#include "Protocol/SimpleNetwork.h"
#include "protocol/gpxe_download.h"

#define BOOTIF_OPTION_SIZE  28  /* STRSIZE("BOOTIF=xx-aa-bb-cc-dd-ee-ff") */

static char ipappend[BOOTIF_OPTION_SIZE];

/*-- is_network_boot -----------------------------------------------------------
 *
 *      Check whether we are booted from the network:
 *        1. Check if we have been PXE booted
 *        2. Check if we have been gPXE booted
 *
 * Results
 *      True/false
 *----------------------------------------------------------------------------*/
bool is_network_boot(void)
{
   EFI_GUID SimpleNetworkProto = EFI_SIMPLE_NETWORK_PROTOCOL_GUID;
   EFI_GUID GpxeDownloadProto = GPXE_DOWNLOAD_PROTOCOL_GUID;
   EFI_HANDLE BootVolume;
   EFI_STATUS Status;
   VOID *Proto;

   Status = get_boot_volume(&BootVolume);
   if (EFI_ERROR(Status)) {
      return false;
   }

   Status = get_protocol_interface(BootVolume, &SimpleNetworkProto, &Proto);
   if (!EFI_ERROR(Status)) {
      return true;
   }

   Status = get_protocol_interface(BootVolume, &GpxeDownloadProto, &Proto);

   return !EFI_ERROR(Status);
}

/*-- get_bootif_option ---------------------------------------------------------
 *
 *      Returns the same string as the pxelinux "ipappend 2" option would have
 *      added to the kernel command line.
 *
 *      The result is a string of the form "BOOTIF=xx-aa-bb-cc-dd-ee-ff", where
 *      xx is the Hardware Type Number of the boot interface (see RFC 1700), and
 *      aa:bb:cc:dd:ee:ff is its MAC address.
 *
 *      NOTE: Unfortunately, gPXE does not support providing the boot interface
 *            mac address. As a result, the BOOTIF= option still has to be
 *            passed manually (via <SHIFT+O>, or by editing the boot.cfg
 *            configuration file) when booting from gPXE.
 *
 * Parameters
 *      OUT bootif: a pointer to the statical BOOTIF string.
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_bootif_option(const char **bootif)
{
   EFI_GUID SimpleNetworkProto = EFI_SIMPLE_NETWORK_PROTOCOL_GUID;
   EFI_SIMPLE_NETWORK *Network;
   EFI_HANDLE BootVolume;
   EFI_STATUS Status;

   Status = get_boot_volume(&BootVolume);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   Status = get_protocol_interface(BootVolume, &SimpleNetworkProto,
                                   (VOID **)&Network);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   snprintf(ipappend, sizeof (ipappend),
            "BOOTIF=%02x-%02x-%02x-%02x-%02x-%02x-%02x",
            (uint32_t)Network->Mode->IfType,
            (uint32_t)Network->Mode->CurrentAddress.Addr[0],
            (uint32_t)Network->Mode->CurrentAddress.Addr[1],
            (uint32_t)Network->Mode->CurrentAddress.Addr[2],
            (uint32_t)Network->Mode->CurrentAddress.Addr[3],
            (uint32_t)Network->Mode->CurrentAddress.Addr[4],
            (uint32_t)Network->Mode->CurrentAddress.Addr[5]);

   *bootif = ipappend;

   return error_efi_to_generic(EFI_SUCCESS);
}
