/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * systab.c -- EFI system table parsing
 */

#include <string.h>
#include "efi_private.h"

/*-- efi_guid_cmp --------------------------------------------------------------
 *
 *      Compare to EFI GUID.
 *
 * Parameters
 *      IN Guid1: pointer to the first GUID
 *      IN Guid2: pointer to the second GUID
 *
 * Results
 *      0 if both GUID's are equal, a non-zero value otherwise.
 *----------------------------------------------------------------------------*/
int efi_guid_cmp(EFI_GUID *guid1, EFI_GUID *guid2)
{
    return memcmp(guid1, guid2, sizeof (EFI_GUID)) ? 1 : 0;
}

/*-- efi_get_system_config_table -----------------------------------------------
 *
 *      Get a configuration table base address.
 *
 * Parameters
 *      IN  guid:  pointer to the configuration GUID
 *      OUT table: pointer to the table base address
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS efi_get_system_config_table(EFI_GUID *guid, void **table)
{
   unsigned int i;

   EFI_ASSERT(st != NULL);

   *table = NULL;
   for (i = 0; i < st->NumberOfTableEntries; i++) {
      if (efi_guid_cmp(&st->ConfigurationTable[i].VendorGuid, guid) == 0) {
         *table = st->ConfigurationTable[i].VendorTable;
         break;
      }
   }

   return (*table == NULL) ? EFI_NOT_FOUND : EFI_SUCCESS;
}

/*-- get_acpi_rsdp -------------------------------------------------------------
 *
 *      Get the ACPI RSDP. The returned RSDP may be ACPI 1.0 or 2.0+ compliant;
 *      preference is given to a 2.0+ table if it exists.
 *
 *      UEFI Specification v2.3 (Section 4.6 "Configuration table") says:
 *      "ACPI 2.0 or newer tables should use EFI_ACPI_TABLE_GUID"
 *
 * Parameters
 *      OUT rsdp: pointer to the RSDP
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_acpi_rsdp(void **rsdp)
{
   EFI_STATUS Status;
   int i;
   EFI_GUID guids[3] = {
      EFI_ACPI_TABLE_GUID,
      EFI_ACPI_20_TABLE_GUID,
      ACPI_10_TABLE_GUID
   };

   Status = EFI_NOT_FOUND;

   for (i = 0; i < 3; i++) {
      if (i > 0 && efi_guid_cmp(&guids[0], &guids[i]) == 0) {
         continue;
      }

      Status = efi_get_system_config_table(&guids[i], rsdp);
      if (Status == EFI_SUCCESS) {
         break;
      }
   }

   return error_efi_to_generic(Status);
}

/*-- get_smbios_table ----------------------------------------------------------
 *
 *      Get the SMBIOS table base address.
 *
 * Parameters
 *      OUT smbios: pointer to the SMBIOS floating pointer
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_smbios_table(void **smbios)
{
   EFI_GUID guid = SMBIOS_TABLE_GUID;
   EFI_STATUS Status;

   Status = efi_get_system_config_table(&guid, smbios);

   return error_efi_to_generic(Status);
}

/*-- get_mps_table -------------------------------------------------------------
 *
 *      Get the MPS floating pointer base address.
 *
 * Parameters
 *      OUT mps: pointer to the MPS table
 *
 * Results
 *      ERR_SUCCESS, or a generic error status.
 *----------------------------------------------------------------------------*/
int get_mps_table(void **mps)
{
   EFI_GUID guid = MPS_TABLE_GUID;
   EFI_STATUS Status;

   Status = efi_get_system_config_table(&guid, mps);

   return error_efi_to_generic(Status);
}
