/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * tftpfile.c --
 *
 *      Provides access to files through the (M)TFTP and PXE Base Code (PXE BC)
 *      Protocols.
 *
 *      Ideally, all files could be loaded using the Load File Protocol,
 *      regardless of the underlying media/transport type. However, the EFI
 *      spec explicitly states that the PXE BC's implementation of LoadFile()
 *      may only be used to discover and retrieve the bootstrap program
 *      (i.e. this bootloader). Other files (e.g. the kernel) must be retrieved
 *      using explicit TFTP calls. That's the purpose of this file.
 *
 *      The following code does not perform a complete PXE exchange. Instead,
 *      the assumption is made that if this is a PXE boot, the PXE BC and MTFTP
 *      modules must have been initialized and the PXE protocol carried out in
 *      order to discover and download this program. Therefore, after proper
 *      sanity checks, the TFTP functionality is accessed directly.
 */

#include <string.h>
#include "efi_private.h"

/*
 * Contrary to the EFI spec, the EDK implementation of Mtftp() will not
 * negotiate the largest block size with the server if the BlockSize argument
 * is NULL. Further, the elilo sources mention that some real firmware
 * implementations timeout when given a NULL BlockSize. Therefore, we pass in
 * the following as an explicit BlockSize request. Beyond avoiding bugs, doing
 * so may also make transfers faster and allow larger files to be
 * downloaded. At worst, the server will simply request a smaller block size.
 */
static const UINTN DEFAULT_BLOCK_SIZE = 4096;

/*-- get_pxe_info --------------------------------------------------------------
 *
 *      Gather info about the PXE BC instance (if any) attached to the given
 *      handle.
 *
 * Parameters
 *      IN  Volume:   handle to the volume from which to load the file
 *      OUT Pxe:      pointer to the PXE BC interface
 *      OUT ServerIp: the IPv4 address of the PXE server
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
static EFI_STATUS get_pxe_info(EFI_HANDLE Volume, EFI_PXE_BASE_CODE **Pxe,
                               EFI_IPv4_ADDRESS *ServerIp)
{
   EFI_GUID PxeBaseCodeProto = EFI_PXE_BASE_CODE_PROTOCOL_GUID;
   EFI_PXE_BASE_CODE_MODE *PxeMode;
   EFI_PXE_BASE_CODE_PACKET *Packet;
   EFI_STATUS Status;

   EFI_ASSERT_PARAM(Pxe != NULL);
   EFI_ASSERT_PARAM(ServerIp != NULL);

   Status = get_protocol_interface(Volume, &PxeBaseCodeProto, (void **)Pxe);
   if (EFI_ERROR(Status)) {
      return Status;
   }
   PxeMode = (*Pxe)->Mode;

   if (!PxeMode->Started) {
      return EFI_NOT_STARTED;
   }

   /*
    * IPv6 isn't supported because the PXE BC doesn't fully support it (for
    * example, EFI_PXE_BASE_CODE_PACKET does not contain an IPv6 packet).
    */
   if (PxeMode->UsingIpv6) {
      return EFI_UNSUPPORTED;
   }

   if (PxeMode->PxeReplyReceived) {
      Packet = &PxeMode->PxeReply;
   } else if (PxeMode->ProxyOfferReceived) {
      Packet = &PxeMode->ProxyOffer;
   } else if (PxeMode->DhcpAckReceived) {
      Packet = &PxeMode->DhcpAck;
   } else {
      return EFI_UNSUPPORTED;
   }

   memcpy(ServerIp, Packet->Dhcpv4.BootpSiAddr, sizeof(*ServerIp));

   return EFI_SUCCESS;
}

/*-- tftp_file_get_size --------------------------------------------------------
 *
 *      Get the size of a file using TFTP.
 *
 *      If downloading from a TFTP server without the 'tsize' option, it is
 *      possible that this could end up downloading the whole file and throwing
 *      its contents away. Get yourself a better TFTP server!
 *
 * Parameters
 *      IN  Volume:   handle to the volume from which to load the file
 *      IN  filepath: the ASCII absolute path of the file to retrieve
 *      OUT FileSize: file size in bytes
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
EFI_STATUS tftp_file_get_size(EFI_HANDLE Volume, const char *filepath,
                              UINTN *FileSize)
{
   EFI_PXE_BASE_CODE *Pxe;
   EFI_IP_ADDRESS ServerIp;
   /* Some firmware doesn't like a NULL BufferPtr in the call to Mtftp(). */
   UINT8 DummyBuf;
   UINT64 Size = 0;
   EFI_STATUS Status;

   EFI_ASSERT_PARAM(filepath != NULL);
   EFI_ASSERT_PARAM(FileSize != NULL);

   Status = get_pxe_info(Volume, &Pxe, &ServerIp.v4);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   efi_set_watchdog_timer(WATCHDOG_DISABLE);

   Status = Pxe->Mtftp(Pxe, EFI_PXE_BASE_CODE_TFTP_GET_FILE_SIZE,
                       &DummyBuf, FALSE, &Size, (UINTN *)&DEFAULT_BLOCK_SIZE,
                       &ServerIp, (UINT8 *)filepath, NULL, TRUE);

   efi_set_watchdog_timer(WATCHDOG_DEFAULT_TIMEOUT);

   /*
    * Some firmware returns EFI_BUFFER_TOO_SMALL even though it successfully
    * managed to get the size.
    */
   if (Status != EFI_BUFFER_TOO_SMALL && EFI_ERROR(Status)) {
      return Status;
   }

   *FileSize = (UINTN)Size;

   return EFI_SUCCESS;
}

/*-- tftp_file_load ------------------------------------------------------------
 *
 *      Load a file into memory using TFTP. UEFI watchdog timer is disabled
 *      during the Mtftp() operation, so it does not trigger and reboot the
 *      platform during large/slow file transfers.
 *
 * Parameters
 *      IN  Volume:   handle to the volume from which to load the file
 *      IN  filepath: the ASCII absolute path of the file to retrieve
 *      IN  callback: routine to be called periodically while the file is being
 *                    loaded
 *      OUT Buffer:   pointer to where to load the file
 *      IN  BufSize:  output buffer size in bytes
 *      OUT BufSize:  number of bytes that have been written into Buffer
 *
 * Results
 *      EFI_SUCCESS, or an UEFI error status.
 *----------------------------------------------------------------------------*/
EFI_STATUS tftp_file_load(EFI_HANDLE Volume, const char *filepath,
                          int (*callback)(size_t), VOID **Buffer,
                          UINTN *BufSize)
{
   EFI_PXE_BASE_CODE *Pxe;
   EFI_IP_ADDRESS ServerIp;
   VOID *Data;
   UINTN Size;
   UINT64 Size64;
   EFI_STATUS Status;
   int error;

   EFI_ASSERT_PARAM(filepath != NULL);
   EFI_ASSERT_PARAM(Buffer != NULL);
   EFI_ASSERT_PARAM(BufSize != NULL);

   Status = tftp_file_get_size(Volume, filepath, &Size);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   Status = get_pxe_info(Volume, &Pxe, &ServerIp.v4);
   if (EFI_ERROR(Status)) {
      return Status;
   }

   Data = sys_malloc(Size);
   if (Data == NULL) {
      return EFI_OUT_OF_RESOURCES;
   }

   Size64 = Size;
   efi_set_watchdog_timer(WATCHDOG_DISABLE);

   Status = Pxe->Mtftp(Pxe, EFI_PXE_BASE_CODE_TFTP_READ_FILE,
                       Data, FALSE, &Size64, (UINTN *)&DEFAULT_BLOCK_SIZE,
                       &ServerIp, (UINT8 *)filepath, NULL, FALSE);

  efi_set_watchdog_timer(WATCHDOG_DEFAULT_TIMEOUT);

   if (EFI_ERROR(Status)) {
      sys_free(Data);
      return Status;
   }

   /*
    * XXX: Use EFI_PXE_BASE_CODE_CALLBACK to call the progress callback each
    * time a packet is received.
    */
   if (callback != NULL) {
      error = callback((size_t)Size64);
      if (error != 0) {
         sys_free(Data);
         return error_generic_to_efi(error);
      }
   }

   *Buffer = Data;
   *BufSize = (UINTN)Size64;

   return EFI_SUCCESS;
}
