/*******************************************************************************
 * Copyright (c) 2008-2011 VMware, Inc.  All rights reserved.
 *
 * This file is part of the esxboot project.
 *
 * esxboot is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation version 2 and no later version.
 *
 * esxboot is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License version 2 for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 ******************************************************************************/

/*
 * isobounce.c --
 *
 *      This program is a simple chainloader for booting off of ISO9660
 *      formatted CDROM media.
 *
 *      EFI only provides a FAT filesystem driver natively. In order to boot
 *      from a CDROM, the EFI firmware looks at a special 'El Torito' boot
 *      entry, which tells it the starting LBA of a FAT filesystem image on the
 *      CDROM. The EFI shell and other EFI applications can access this FAT
 *      filesystem image, but they cannot access the rest of the contents of
 *      the CDROM.
 *
 *      This program gets around this limitation of the EFI firmware by
 *      installing an ISO9660 filesystem driver, then loading and transferring
 *      control to a second EFI application located in the ISO9660 filesystem.
 *      You put this program in a tiny FAT-formatted image that is pointed to
 *      by an El Torito boot entry, and you put your regular EFI bootloader,
 *      kernel images, initrds, kernel modules, and everything else in the
 *      ISO9660 filesystem. This is particularly useful if the kernel or
 *      modules can also be loaded by ISOLINUX (or another BIOS bootloader),
 *      because this way you do not end up with two copies of everything on the
 *      CD (in two different filesystems).
 */

#include <efiutils.h>
#include <bootlib.h>

#ifndef ISO9660_DRIVER
   #if defined(__x86_64__)
      #define ISO9660_DRIVER L"\\EFI\\DRIVERS\\ISO9660x64.EFI"
   #else
      #define ISO9660_DRIVER L"\\EFI\\DRIVERS\\ISO9660IA32.EFI"
   #endif
#endif

#ifndef NEXT_LOADER
   #if defined(__x86_64__)
      #define NEXT_LOADER L"EFI\\BOOT\\BOOTx64.EFI"
   #else
      #define NEXT_LOADER L"EFI\\BOOT\\BOOTIA32.EFI"
   #endif
#endif

int main(UNUSED_PARAM(int argc), UNUSED_PARAM(char **argv))
{
   EFI_HANDLE BootVolume, CdromDevice;
   EFI_HANDLE DriverImageHandle[2] = {NULL, NULL};
   EFI_STATUS Status, ChildStatus;

   EFI_ASSERT(bs != NULL);
   EFI_ASSERT_FIRMWARE(bs->ConnectController != NULL);
   EFI_ASSERT_FIRMWARE(bs->DisconnectController != NULL);

   /* Locate and load the ISO9660 driver */
   Status = get_boot_volume(&BootVolume);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   Status = image_load(BootVolume, ISO9660_DRIVER, NULL, 0,
                       &DriverImageHandle[0], &ChildStatus);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }
   if (EFI_ERROR(ChildStatus)) {
      return error_efi_to_generic(ChildStatus);
   }

   /* Disconnect all drivers from the CDROM, then connect the ISO9660 driver. */
   Status = get_boot_device(&CdromDevice);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   Status = bs->DisconnectController(CdromDevice, NULL, NULL);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   Status = bs->ConnectController(CdromDevice, DriverImageHandle, NULL, FALSE);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }

   /* Chainload the bootloader */
   Status = image_load(CdromDevice, NEXT_LOADER, NULL, 0, NULL, &ChildStatus);
   if (EFI_ERROR(Status)) {
      return error_efi_to_generic(Status);
   }
   if (EFI_ERROR(ChildStatus)) {
      return error_efi_to_generic(ChildStatus);
   }

   return error_efi_to_generic(EFI_SUCCESS);
}
